# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

"""
Mounts NCCH containers, creating a virtual filesystem of decrypted sections.
"""

import logging
from errno import ENOENT
from stat import S_IFDIR, S_IFREG
from sys import argv
from typing import TYPE_CHECKING

from pyctr.crypto import load_seeddb
from pyctr.type.ncch import NCCHReader, NCCHSection

from . import _common as _c
# _common imports these from fusepy, and prints an error if it fails; this allows less duplicated code
from ._common import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context, get_time, load_custom_boot9, \
    realpath
from .exefs import ExeFSMount
from .romfs import RomFSMount

if TYPE_CHECKING:
    from typing import Dict


class NCCHContainerMount(LoggingMixIn, Operations):
    fd = 0
    romfs_fuse = None
    exefs_fuse = None

    def __init__(self, reader: 'NCCHReader', g_stat: dict):
        self.files: Dict[str, NCCHSection] = {}

        # get status change, modify, and file access times
        self.g_stat = g_stat

        self.reader = reader

    def __del__(self, *args):
        try:
            self.reader.close()
        except AttributeError:
            pass

    destroy = __del__

    def init(self, path, _setup_romfs=True):
        decrypted_filename = '/decrypted.' + ('cxi' if self.reader.flags.executable else 'cfa')

        self.files[decrypted_filename] = NCCHSection.FullDecrypted
        self.files['/ncch.bin'] = NCCHSection.Header

        if NCCHSection.ExtendedHeader in self.reader.sections:
            self.files['/extheader.bin'] = NCCHSection.ExtendedHeader

        if NCCHSection.Logo in self.reader.sections:
            self.files['/logo.bin'] = NCCHSection.Logo

        if NCCHSection.Plain in self.reader.sections:
            self.files['/plain.bin'] = NCCHSection.Plain

        if NCCHSection.ExeFS in self.reader.sections:
            self.files['/exefs.bin'] = NCCHSection.ExeFS
            self.exefs_fuse = ExeFSMount(self.reader.exefs, g_stat=self.g_stat, decompress_code=True)
            self.exefs_fuse.init(path)

        if NCCHSection.RomFS in self.reader.sections:
            self.files['/romfs.bin'] = NCCHSection.RomFS
            self.romfs_fuse = RomFSMount(self.reader.romfs, g_stat=self.g_stat)

    @_c.ensure_lower_path
    def getattr(self, path, fh=None):
        if path.startswith('/exefs/'):
            return self.exefs_fuse.getattr(_c.remove_first_dir(path), fh)
        elif path.startswith('/romfs/'):
            return self.romfs_fuse.getattr(_c.remove_first_dir(path), fh)
        uid, gid, pid = fuse_get_context()
        if path in {'/', '/romfs', '/exefs'}:
            st = {'st_mode': (S_IFDIR | 0o777), 'st_nlink': 2}
        elif path in self.files:
            st = {'st_mode': (S_IFREG | 0o666), 'st_size': self.reader.sections[self.files[path]].size, 'st_nlink': 1}
        else:
            raise FuseOSError(ENOENT)
        return {**st, **self.g_stat, 'st_uid': uid, 'st_gid': gid}

    def open(self, path, flags):
        self.fd += 1
        return self.fd

    @_c.ensure_lower_path
    def readdir(self, path, fh):
        if path.startswith('/exefs'):
            yield from self.exefs_fuse.readdir(_c.remove_first_dir(path), fh)
        elif path.startswith('/romfs'):
            yield from self.romfs_fuse.readdir(_c.remove_first_dir(path), fh)
        elif path == '/':
            yield from ('.', '..')
            yield from (x[1:] for x in self.files)
            if self.exefs_fuse is not None:
                yield 'exefs'
            if self.romfs_fuse is not None:
                yield 'romfs'

    @_c.ensure_lower_path
    def read(self, path, size, offset, fh):
        if path.startswith('/exefs/'):
            return self.exefs_fuse.read(_c.remove_first_dir(path), size, offset, fh)
        elif path.startswith('/romfs/'):
            return self.romfs_fuse.read(_c.remove_first_dir(path), size, offset, fh)

        section = self.files[path]
        with self.reader.open_raw_section(section) as f:
            f.seek(offset)
            return f.read(size)

    @_c.ensure_lower_path
    def statfs(self, path):
        if path.startswith('/exefs/'):
            return self.exefs_fuse.statfs(_c.remove_first_dir(path))
        elif path.startswith('/romfs/'):
            return self.romfs_fuse.statfs(_c.remove_first_dir(path))
        else:
            return {'f_bsize': 4096, 'f_frsize': 4096, 'f_blocks': self.reader.content_size // 4096, 'f_bavail': 0,
                    'f_bfree': 0, 'f_files': len(self.files)}


def main(prog: str = None, args: list = None):
    from argparse import ArgumentParser
    if args is None:
        args = argv[1:]
    parser = ArgumentParser(prog=prog, description='Mount Nintendo 3DS NCCH containers.',
                            parents=(_c.default_argp, _c.ctrcrypto_argp, _c.seeddb_argp,
                                     _c.main_args('ncch', 'NCCH file')))
    parser.add_argument('--dec', help='assume contents are decrypted', action='store_true')

    a = parser.parse_args(args)
    opts = dict(_c.parse_fuse_opts(a.o))

    if a.do:
        logging.basicConfig(level=logging.DEBUG, filename=a.do)

    ncch_stat = get_time(a.ncch)

    load_custom_boot9(a.boot9)

    if a.seeddb:
        load_seeddb(a.seeddb)

    with NCCHReader(a.ncch, dev=a.dev, assume_decrypted=a.dec, seed=a.seed) as r:
        mount = NCCHContainerMount(reader=r, g_stat=ncch_stat)
        if _c.macos or _c.windows:
            opts['fstypename'] = 'NCCH'
            if _c.macos:
                display = f'{r.partition_id.upper()}; {r.product_code}'
                try:
                    title = r.exefs.icon.get_app_title()
                    if title.short_desc != 'unknown':
                        display += '; ' + title.short_desc
                except:
                    pass
                opts['volname'] = f'NCCH Container ({display})'
            elif _c.windows:
                # volume label can only be up to 32 chars
                try:
                    title = r.exefs.icon.get_app_title().short_desc
                    if len(title) > 26:
                        title = title[0:25] + '\u2026'  # ellipsis
                    display = title
                except:
                    display = r.partition_id.upper()
                opts['volname'] = f'NCCH ({display})'
        FUSE(mount, a.mount_point, foreground=a.fg or a.d, ro=True, nothreads=True, debug=a.d,
             fsname=realpath(a.ncch).replace(',', '_'), **opts)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

types = {
    'cci': {
        'name': 'CTR Cart Image',
        'info': '".3ds", ".cci"'
    },
    'cdn': {
        'name': 'CDN contents',
        'info': '"cetk", "tmd", and contents'
    },
    'cia': {
        'name': 'CTR Importable Archive',
        'info': '".cia"'
    },
    'exefs': {
        'name': 'Executable Filesystem',
        'info': '".exefs", "exefs.bin"'
    },
    'nandctr': {
        'name': 'Nintendo 3DS NAND backup',
        'info': '"nand.bin"'
    },
    'nandtwl': {
        'name': 'Nintendo DSi NAND backup',
        'info': '"nand_dsi.bin"'
    },
    'nandhac': {
        'name': 'Nintendo Switch NAND backup',
        'info': '"rawnand.bin"'
    },
    'nandbb': {
        'name': 'iQue Player NAND backup',
        'info': '"nand.bin"'
    },
    'ncch': {
        'name': 'NCCH',
        'info': '".cxi", ".cfa", ".ncch", ".app"'
    },
    'romfs': {
        'name': 'Read-only Filesystem',
        'info': '".romfs", "romfs.bin"'
    },
    'sd': {
        'name': 'SD Card Contents',
        'info': '"Nintendo 3DS" from SD'
    },
    'sdtitle': {
        'name': 'Installed SD Title Contents',
        'info': '"*.tmd" and "*.app" files'
    },
    'srl': {
        'name': 'Nintendo DS ROM image',
        'info': '".nds", ".srl"'
    },
    'threedsx': {
        'name': '3DSX Homebrew',
        'info': '".3dsx"'
    },
}

aliases = {
    '3ds': 'cci',
    '3dsx': 'threedsx',
    'app': 'ncch',
    'csu': 'cci',
    'cxi': 'ncch',
    'cfa': 'ncch',
    'nand': 'nandctr',
    'nanddsi': 'nandtwl',
    'nandique': 'nandbb',
    'nandswitch': 'nandhac',
    'nandnx': 'nandhac',
    'nds': 'srl',
}

categories = {
    'Nintendo 3DS': ['cci', 'cdn', 'cia', 'exefs', 'nandctr', 'ncch', 'romfs', 'sd', 'sdtitle', 'threedsx'],
    'Nintendo DS / DSi': ['nandtwl', 'srl'],
    'Nintendo Switch': ['nandhac'],
    'iQue Player': ['nandbb']
}

# this will add the "Use developer-unit keys" option to Advanced options in the gui
supports_dev_keys = ['cci', 'cdn', 'cia', 'nandctr', 'ncch', 'sd', 'sdtitle']


def get_type_info(mount_type):
    return types[aliases.get(mount_type, mount_type)]



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import tkinter as tk
from typing import TYPE_CHECKING

from .base import WizardBase

if TYPE_CHECKING:
    from .. import WizardContainer


class ThreeDSXSetup(WizardBase):
    def __init__(self, parent: 'tk.BaseWidget' = None, *, wizardcontainer: 'WizardContainer'):
        super().__init__(parent, wizardcontainer=wizardcontainer)

        def callback(*_):
            main_file = self.main_textbox_var.get().strip()
            self.wizardcontainer.set_next_enabled(main_file)

        main_container, main_textbox, main_textbox_var = self.make_file_picker('Select the 3dsx file:',
                                                                               'Select 3dsx file')
        main_container.pack(fill=tk.X, expand=True)

        self.main_textbox_var = main_textbox_var

        main_textbox_var.trace_add('write', callback)

        self.set_header_suffix('3DSX')

    def next_pressed(self):
        main_file = self.main_textbox_var.get().strip()

        args = ['threedsx', main_file]

        self.wizardcontainer.show_mount_point_selector('3DSX', args)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

"""
Mounts SD contents under `/Nintendo 3DS`, creating a virtual filesystem with decrypted contents. movable.sed required.
"""

import logging
import os
from errno import EPERM, EACCES
from os.path import basename, dirname, isdir
from sys import exit, argv
from threading import Lock
from typing import TYPE_CHECKING

from pyctr.crypto import CryptoEngine, Keyslot

from . import _common as _c
# _common imports these from fusepy, and prints an error if it fails; this allows less duplicated code
from ._common import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context, realpath

if _c.windows:
    from ctypes import c_wchar_p, pointer, c_ulonglong, windll, wintypes

if TYPE_CHECKING:
    from typing import BinaryIO, Dict, Optional, Tuple


class SDFilesystemMount(LoggingMixIn, Operations):

    @_c.ensure_lower_path
    def path_to_iv(self, path):
        return CryptoEngine.sd_path_to_iv(path[self.root_len + 33:])

    def fd_to_fileobj(self, path, mode, fd):
        fh = open(fd, mode, buffering=0)
        lock = Lock()
        if not (basename(path).startswith('.') or 'nintendo dsiware' in path.lower() or dirname(path) == self.root):
            fh_enc = self.crypto.create_ctr_io(Keyslot.SD, fh, self.path_to_iv(path))
            fh_group = (fh_enc, fh, lock)
        else:
            fh_group = (fh, None, lock)
        self.fds[fd] = fh_group
        return fd

    def __init__(self, sd_dir: str, movable: bytes, dev: bool = False, readonly: bool = False, boot9: str = None):
        self.crypto = CryptoEngine(boot9=boot9, dev=dev)

        # only allows one create/open/release operation at a time
        self.global_lock = Lock()

        # each fd contains a tuple with a file object, a base file object (for encrypted files),
        #   and a thread lock to prevent two read or write operations from screwing with eachother
        self.fds: 'Dict[int, Tuple[BinaryIO, Optional[BinaryIO], Lock]]' = {}

        self.crypto.setup_sd_key(movable)
        self.root_dir = self.crypto.id0.hex()

        if not isdir(sd_dir + '/' + self.root_dir):
            exit(f'Could not find ID0 {self.root_dir} in the SD directory.')

        print('ID0:', self.root_dir)
        print('Key:', self.crypto.keygen(Keyslot.SD).hex())

        self.root = realpath(sd_dir + '/' + self.root_dir)
        self.root_len = len(self.root)

        self.readonly = readonly

    # noinspection PyMethodOverriding
    def __call__(self, op, path, *args):
        return super().__call__(op, self.root + path, *args)

    def access(self, path, mode):
        if not os.access(path, mode):
            raise FuseOSError(EACCES)

    @_c.raise_on_readonly
    def chmod(self, path, mode):
        os.chmod(path, mode)

    @_c.raise_on_readonly
    def chown(self, path, *args, **kwargs):
        if not _c.windows:
            os.chown(path, *args, **kwargs)

    @_c.raise_on_readonly
    def create(self, path, mode, fi=None):
        # prevent another create/open/release from interfering
        with self.global_lock:
            fd = os.open(path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, mode)
            return self.fd_to_fileobj(path, 'wb+', fd)

    @_c.raise_on_readonly
    def flush(self, path, fh):
        fd, _, lock = self.fds[fh]
        with lock:
            fd.flush()

    def getattr(self, path, fh=None):
        st = os.lstat(path)
        uid, gid, _ = fuse_get_context()
        res = {key: getattr(st, key) for key in ('st_atime', 'st_ctime', 'st_mode', 'st_mtime', 'st_nlink', 'st_size',
                                                 'st_flags') if hasattr(st, key)}
        res['st_uid'] = st.st_uid if st.st_uid != 0 else uid
        res['st_gid'] = st.st_gid if st.st_gid != 0 else gid
        return res

    def link(self, target, source):
        return os.link(source, target)

    @_c.raise_on_readonly
    def mkdir(self, path, mode):
        os.mkdir(path, mode)

    @_c.raise_on_readonly
    def mknod(self, path, mode, dev):
        if not _c.windows:
            os.mknod(path, mode, dev)

    # open = os.open
    def open(self, path, flags):
        # prevent another create/open/release from interfering
        with self.global_lock:
            fd = os.open(path, flags)
            return self.fd_to_fileobj(path, 'rb+', fd)

    def read(self, path, size, offset, fh):
        fd, _, lock = self.fds[fh]

        # acquire lock to prevent another read/write from messing with this operation
        with lock:
            fd.seek(offset)
            return fd.read(size)

    def readdir(self, path, fh):
        yield from ('.', '..')

        # due to DSiWare exports having unique crypto that is a pain to handle, this hides it to prevent misleading
        #   users into thinking that the files are decrypted.
        ld = (d for d in os.listdir(path) if not d.lower() == 'nintendo dsiware')

        if _c.windows:
            # I should figure out how to mark hidden files, if possible
            yield from (d for d in ld if not d.startswith('.'))
        else:
            yield from ld

    readlink = os.readlink

    def release(self, path, fh):
        # prevent another create/open/release from interfering
        with self.global_lock:
            fd_group = self.fds[fh]
            # prevent use of the handle while cleaning up, or closing while in use
            with fd_group[2]:
                fd_group[0].close()
                try:
                    fd_group[1].close()
                except AttributeError:
                    # unencrypted files have the second item set to None
                    pass
                del self.fds[fh]

    @_c.raise_on_readonly
    def rename(self, old, new):
        # renaming's too difficult. just copy the file to the name you want if you really need it.
        raise FuseOSError(EPERM)

    @_c.raise_on_readonly
    def rmdir(self, path):
        os.rmdir(path)

    # noinspection PyPep8Naming
    def statfs(self, path):
        if _c.windows:
            lpSectorsPerCluster = c_ulonglong(0)
            lpBytesPerSector = c_ulonglong(0)
            lpNumberOfFreeClusters = c_ulonglong(0)
            lpTotalNumberOfClusters = c_ulonglong(0)
            ret = windll.kernel32.GetDiskFreeSpaceW(c_wchar_p(path), pointer(lpSectorsPerCluster),
                                                    pointer(lpBytesPerSector),
                                                    pointer(lpNumberOfFreeClusters),
                                                    pointer(lpTotalNumberOfClusters))
            if not ret:
                raise WindowsError
            free_blocks = lpNumberOfFreeClusters.value * lpSectorsPerCluster.value
            result = {'f_bavail': free_blocks,
                      'f_bfree': free_blocks,
                      'f_bsize': lpBytesPerSector.value,
                      'f_frsize': lpBytesPerSector.value,
                      'f_blocks': lpTotalNumberOfClusters.value * lpSectorsPerCluster.value,
                      'f_namemax': wintypes.MAX_PATH}
            return result
        else:
            stv = os.statvfs(path)
            # f_flag causes python interpreter crashes in some cases. i don't get it.
            return {key: getattr(stv, key) for key in ('f_bavail', 'f_bfree', 'f_blocks', 'f_bsize', 'f_favail',
                                                       'f_ffree', 'f_files', 'f_frsize', 'f_namemax')}

    def symlink(self, target, source):
        return os.symlink(source, target)

    def truncate(self, path, length, fh=None):
        try:
            fd, _, lock = self.fds[fh]
            # acquire lock to prevent another read/write from messing with this operation
            with lock:
                fd.truncate(length)

        except KeyError:  # in case this is not an already open file
            with open(path, 'rb+') as f:
                f.truncate(length)

    @_c.raise_on_readonly
    def unlink(self, path):
        os.unlink(path)

    @_c.raise_on_readonly
    def utimens(self, path, *args, **kwargs):
        os.utime(path, *args, **kwargs)

    @_c.raise_on_readonly
    def write(self, path, data, offset, fh):
        fd, _, lock = self.fds[fh]

        # acquire lock to prevent another read/write from messing with this operation
        with lock:
            fd.seek(offset)
            return fd.write(data)


def main(prog: str = None, args: list = None):
    from argparse import ArgumentParser
    if args is None:
        args = argv[1:]
    parser = ArgumentParser(prog=prog, description='Mount Nintendo 3DS SD card contents.',
                            parents=(_c.default_argp, _c.readonly_argp, _c.ctrcrypto_argp,
                                     _c.main_args(
                                         'sd_dir', "path to folder with SD contents (on SD: /Nintendo 3DS)")))
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--movable', metavar='MOVABLESED', help='path to movable.sed')
    group.add_argument('--sd-key', metavar='SDKEY', help='SD key as hexstring')

    a = parser.parse_args(args)
    opts = dict(_c.parse_fuse_opts(a.o))

    if a.do:
        logging.basicConfig(level=logging.DEBUG, filename=a.do)

    if a.movable:
        with open(a.movable, 'rb') as f:
            movable = f.read(0x140)
    else:
        movable = bytes.fromhex(a.sd_key)

    mount = SDFilesystemMount(sd_dir=a.sd_dir, movable=movable, dev=a.dev, readonly=a.ro, boot9=a.boot9)
    if _c.macos or _c.windows:
        opts['fstypename'] = 'SDCard'
        if _c.macos:
            opts['volname'] = f'Nintendo 3DS SD Card ({mount.root_dir})'
            # this fixes an issue with creating files through Finder
            # a better fix would probably be to support xattrs properly, but given the kind of data and filesystem
            #   that this code would interact with, it's not really useful.
            opts['noappledouble'] = True
        else:
            # windows
            opts['volname'] = f'Nintendo 3DS SD Card ({mount.root_dir[0:8]}…)'
            opts['case_insensitive'] = False
    FUSE(mount, a.mount_point, foreground=a.fg or a.d, ro=a.ro, debug=a.d,
         fsname=realpath(a.sd_dir).replace(',', '_'), **opts)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import tkinter as tk
import tkinter.ttk as ttk
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable


class OutputViewer(ttk.Frame):
    def __init__(self, parent: 'tk.BaseWidget' = None, *, output: 'Iterable[str]'):
        super().__init__(parent)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=0)

        scrollbar = ttk.Scrollbar(self, orient=tk.VERTICAL)
        scrollbar.grid(row=0, column=1, sticky=tk.NSEW)

        textarea = tk.Text(self, wrap='word', yscrollcommand=scrollbar.set)
        textarea.grid(row=0, column=0, sticky=tk.NSEW)

        scrollbar.configure(command=textarea.yview)

        for line in output:
            textarea.insert(tk.END, line + '\n')

        textarea.see(tk.END)
        textarea.configure(state=tk.DISABLED)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

from ctypes import windll
from os.path import isdir, expandvars
from sys import stderr
import winreg
from argparse import ArgumentParser

SendMessageTimeoutW = windll.user32.SendMessageTimeoutW

HWND_BROADCAST = 0xFFFF
WM_SETTINGCHANGE = 0x001A
SMTO_NORMAL = 0


def refresh_environment():
    res = SendMessageTimeoutW(HWND_BROADCAST, WM_SETTINGCHANGE, 0, 'Environment', SMTO_NORMAL, 10, 0)
    if not res:
        print('Failed to tell explorer about the updated environment.')
        print('SendMessageTimeoutW:', res)


def do(op: str, mypath: str, allusers: bool):
    access = winreg.KEY_READ
    if op in {'add', 'remove'}:
        access |= winreg.KEY_WRITE
    if allusers:
        key = winreg.HKEY_LOCAL_MACHINE
        sub_key = r'SYSTEM\CurrentControlSet\Control\Session Manager\Environment'
    else:
        key = winreg.HKEY_CURRENT_USER
        sub_key = 'Environment'
    try:
        k = winreg.OpenKey(key, sub_key, 0, access)
    except PermissionError as e:
        print('This program needs to be run as administrator to edit environment variables for all users.', file=stderr)
        print(f'{type(e).__name__}: {e}', file=stderr)
        return
    value, keytype = winreg.QueryValueEx(k, 'Path')

    paths: 'list[str]' = value.strip(';').split(';')
    update = False
    if op == 'add':
        if mypath not in paths:
            paths.append(mypath)
            update = True
        else:
            print('Already in Path, not adding')
    elif op == 'remove':
        try:
            paths.remove(mypath)
            update = True
        except ValueError:
            print('Not in Path')
    elif op == 'list':
        for path in paths:
            print(path)
    elif op == 'check':
        for idx, path in enumerate(paths):
            print(f'{idx}: {path}')
            expanded = expandvars(path)
            if expanded != path:
                print(f'  {expanded}')
            if not isdir(expanded):
                print('  not a directory')

    if update:
        winreg.SetValueEx(k, 'Path', 0, keytype, ';'.join(paths))
    winreg.CloseKey(k)
    if update:
        refresh_environment()


if __name__ == '__main__':
    parser = ArgumentParser(description=r'Modify Path environment variable')
    parser.add_argument('-allusers', help='Modify HKLM (Path for all users), defaults to HKCU (Path for current user)', action='store_true')
    opers = parser.add_mutually_exclusive_group(required=True)
    opers.add_argument('-add', metavar='PATH', help='Add path')
    opers.add_argument('-remove', metavar='PATH', help='Remove path')
    opers.add_argument('-list', help='List paths', action='store_true')
    opers.add_argument('-check', help='Check paths', action='store_true')

    args = parser.parse_args()

    if args.add:
        do('add', args.add, args.allusers)
    elif args.remove:
        do('remove', args.remove, args.allusers)
    elif args.list:
        do('list', '', args.allusers)
    elif args.check:
        do('check', '', args.allusers)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import logging
import os
from collections import defaultdict
from errno import ENOENT, EROFS
from inspect import cleandoc
from stat import S_IFDIR, S_IFREG
from sys import argv, exit
from typing import TYPE_CHECKING
from zlib import crc32

from haccrypto.crypto import XTSN, parse_biskeydump
from ._common import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context, get_time, realpath, basename
from . import _common as _c

if TYPE_CHECKING:
    from typing import BinaryIO, List, TextIO

bis_key_ids = defaultdict(lambda: -1, {
    'PRODINFO': 0,
    'PRODINFOF': 0,
    'SAFE': 1,
    'SYSTEM': 2,
    'USER': 3
})


class HACNandImageMount(LoggingMixIn, Operations):
    fd = 0

    def __init__(self, nand_fp: 'BinaryIO', g_stat: dict, keys: str, readonly: bool = False, emummc: int = None,
                 partition: str = None):
        self.base_addr = emummc or 0
        self.readonly = readonly
        self.g_stat = g_stat

        bis_keys = parse_biskeydump(keys)
        self.crypto: List[XTSN] = [None] * 4
        for x in range(4):
            try:
                self.crypto[x] = XTSN(*bis_keys[x])
            except TypeError:
                print(f"Couldn't get BIS KEY {x}. The associated partition(s) will not be available.")
                self.crypto[x] = None

        self.files = {}
        if partition:
            # we can probably assume this is a real file, and therefore can seek to the end this way...
            part_name = partition.upper()
            nand_fp.seek(0, 2)
            size = nand_fp.tell()
            self.files[f'/{part_name.lower()}.img'] = {'real_filename': part_name + '.img',
                                                       'bis_key': bis_key_ids[part_name],
                                                       'start': 0, 'end': size}
        else:
            nand_fp.seek(0x200 + self.base_addr)
            gpt_header = nand_fp.read(0x5C)
            if gpt_header[0:8] != b'EFI PART':
                exit('GPT header magic not found.')

            header_to_hash = gpt_header[0:0x10] + b'\0\0\0\0' + gpt_header[0x14:]
            crc_expected = int.from_bytes(gpt_header[0x10:0x14], 'little')
            crc_got = crc32(header_to_hash) & 0xFFFFFFFF
            if crc_got != crc_expected:
                exit(f'GPT header crc32 mismatch (expected {crc_expected:08x}, got {crc_got:08x})')

            gpt_backup_header_location = int.from_bytes(gpt_header[0x20:0x28], 'little')
            # check if the backup header exists
            nand_fp.seek(gpt_backup_header_location * 0x200 + self.base_addr)
            gpt_backup_header = nand_fp.read(0x200)
            if gpt_backup_header[0:8] != b'EFI PART':
                if emummc:
                    # it seems sometimes the backup header is not found in an emummc image
                    print('Warning: GPT backup header not found.')
                else:
                    exit('GPT backup header not found. This likely means an incomplete backup.')

            gpt_part_start = int.from_bytes(gpt_header[0x48:0x50], 'little')
            gpt_part_count = int.from_bytes(gpt_header[0x50:0x54], 'little')
            gpt_part_entry_size = int.from_bytes(gpt_header[0x54:0x58], 'little')

            nand_fp.seek(gpt_part_start * 0x200 + self.base_addr)
            gpt_part_full_raw = nand_fp.read(gpt_part_count * gpt_part_entry_size)
            gpt_part_crc_expected = int.from_bytes(gpt_header[0x58:0x5C], 'little')
            gpt_part_crc_got = crc32(gpt_part_full_raw) & 0xFFFFFFFF
            if gpt_part_crc_got != gpt_part_crc_expected:
                exit(f'GPT Partition table crc32 mismatch '
                     f'(expected {gpt_part_crc_expected:08x}, got {gpt_part_crc_got:08x})')
            gpt_parts_raw = [gpt_part_full_raw[i:i + gpt_part_entry_size] for i in range(0, len(gpt_part_full_raw),
                                                                                         gpt_part_entry_size)]
            for part in gpt_parts_raw:
                name = part[0x38:].decode('utf-16le').rstrip('\0')
                # check if we have a key for this partition
                if self.crypto[bis_key_ids[name]] is not None:
                    self.files[f'/{name.lower()}.img'] = {'real_filename': name + '.img', 'bis_key': bis_key_ids[name],
                                                          'start': int.from_bytes(part[0x20:0x28], 'little') * 0x200,
                                                          'end': (int.from_bytes(part[0x28:0x30], 'little') + 1) * 0x200}

        self.f = nand_fp

    def __del__(self, *args):
        try:
            self.f.close()
        except AttributeError:
            pass

    destroy = __del__

    def flush(self, path, fh):
        return self.f.flush()

    @_c.ensure_lower_path
    def getattr(self, path: str, fh=None):
        uid, gid, pid = fuse_get_context()
        if path == '/':
            st = {'st_mode': (S_IFDIR | 0o777), 'st_nlink': 2}
        elif path in self.files:
            p = self.files[path]
            st = {'st_mode': (S_IFREG | 0o666),
                  'st_size': p['end'] - p['start'], 'st_nlink': 1}
        else:
            raise FuseOSError(ENOENT)
        return {**st, **self.g_stat, 'st_uid': uid, 'st_gid': gid}

    def open(self, path: str, flags):
        self.fd += 1
        return self.fd

    @_c.ensure_lower_path
    def readdir(self, path: str, fh):
        yield from ('.', '..')
        yield from (x['real_filename'] for x in self.files.values())

    @_c.ensure_lower_path
    def read(self, path: str, size: int, offset: int, fh):
        fi = self.files[path]
        real_offset: int = fi['start'] + offset

        if fi['start'] + offset > fi['end']:
            return b''
        if offset + size > fi['end']:
            size = fi['end'] - offset

        if fi['bis_key'] >= 0:
            before = offset % 16
            after = (offset + size) % 16
            if after:
                after = 16 - after
            aligned_real_offset = real_offset - before
            aligned_offset = offset - before
            size = before + size
            self.f.seek(aligned_real_offset + self.base_addr)
            xtsn = self.crypto[fi['bis_key']]
            return xtsn.decrypt(self.f.read(size + after), 0, 0x4000, aligned_offset)[before:size]

        else:
            self.f.seek(real_offset + self.base_addr)
            return self.f.read(size)

    @_c.ensure_lower_path
    def write(self, path: str, data: bytes, offset: int, fh):
        if self.readonly:
            raise FuseOSError(EROFS)

        fi = self.files[path]
        real_offset: int = fi['start'] + offset
        real_len = len(data)

        if fi['start'] + offset > fi['end']:
            # not writing past the file size
            return real_len

        if real_offset + real_len > fi['end']:
            data = data[:-((real_offset + real_len) - (fi['end'] - fi['start']))]

        if fi['bis_key'] >= 0:
            before = offset % 16
            after = (offset + real_len) % 16
            aligned_offset = offset - before
            aligned_real_offset = real_offset - before
            if after:
                # this sucks...
                new_after = 16 - after
                last_block_ending = self.read(path, new_after, offset + len(data), 0)
            else:
                last_block_ending = b''

            if before:
                first_block_beginning = self.read(path, before, offset - before, 0)
            else:
                first_block_beginning = b''

            self.f.seek(aligned_real_offset + self.base_addr)
            xtsn = self.crypto[fi['bis_key']]
            to_encrypt = b''.join((first_block_beginning, data, last_block_ending))
            self.f.write(xtsn.encrypt(to_encrypt, 0, 0x4000, aligned_offset))

        else:
            self.f.seek(real_offset + self.base_addr)
            self.f.write(data)

        return real_len

    # TODO: get the real nand size, instead of hard-coding it
    @_c.ensure_lower_path
    def statfs(self, path: str):
        return {'f_bsize': 4096, 'f_frsize': 4096, 'f_blocks': 0x747C00000 // 4096, 'f_bavail': 0, 'f_bfree': 0,
                'f_files': len(self.files)}


def main(prog: str = None, args: list = None):
    from argparse import ArgumentParser, RawDescriptionHelpFormatter
    if args is None:
        args = argv[1:]
    parser = ArgumentParser(prog=prog, description='Mount Nintendo Switch NAND images.',
                            parents=(_c.default_argp, _c.readonly_argp,
                                     _c.main_args('image', 'NAND image or encrypted partition')),
                            formatter_class=RawDescriptionHelpFormatter,
                            epilog=cleandoc('''
                            partitions:
                              PRODINFO  (BIS Key 0)
                              PRODINFOF (BIS Key 0)
                              SAFE      (BIS Key 1)
                              SYSTEM    (BIS Key 2)
                              USER      (BIS KEY 3)
                            '''))
    parser.add_argument('--keys', help='keys file from biskeydump or Lockpick_RCM',
                        default=os.path.join(os.path.expanduser('~'), '.switch', 'prod.keys'))
    parser.add_argument('-S', '--split-files', help='treat as part of a split file', action='store_true')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-e', '--emummc', help='is an emuMMC image', action='store_const', const=0x800000)
    group.add_argument('-R', '--raw-emummc', help='is an SD raw emuMMC image', action='store_const', const=0x1800000)
    group.add_argument('--partition', metavar='PARTNAME', help='is a single encrypted partition')

    a = parser.parse_args(args)
    opts = dict(_c.parse_fuse_opts(a.o))

    if a.do:
        logging.basicConfig(level=logging.DEBUG, filename=a.do)

    def do_thing(f: 'BinaryIO', k: 'TextIO', nand_stat: dict):
        mount = HACNandImageMount(nand_fp=f, g_stat=nand_stat, keys=k.read(), readonly=a.ro,
                                  emummc=a.emummc or a.raw_emummc, partition=a.partition)
        if _c.macos or _c.windows:
            opts['fstypename'] = 'HACFS'
            if _c.macos:
                opts['volname'] = f'Nintendo Switch NAND ({basename(a.image)})'
            elif _c.windows:
                # volume label can only be up to 32 chars
                opts['volname'] = 'Nintendo Switch NAND'
        FUSE(mount, a.mount_point, foreground=a.fg or a.d, ro=a.ro, nothreads=True, debug=a.d,
             fsname=realpath(a.image).replace(',', '_'), **opts)

    mode = 'rb' if a.ro else 'r+b'

    with open(a.keys, 'r', encoding='utf-8') as k:
        # get absolute path to filename
        # (works around issues related to current directory when forking into the background)
        image = os.path.abspath(a.image)
        if a.split_files:
            # make sure the ending is an integer
            try:
                int(image[-2:])
            except ValueError:
                exit('Could not find a part number at the end of the filename.\n'
                     'A multi-part Nintendo Switch NAND backup should have filenames in the format of '
                     '"filename.bin.XX", where XX is the part number.')

            # try to find all the parts, starting with 00
            base = image[:-2]
            count = 0
            while True:
                if os.path.isfile(base + format(count, '02')):
                    count += 1
                else:
                    break

            if count == 0:
                exit('Could not find the first part of the multi-part backup.')

            handler = _c.SplitFileHandler((base + format(x, '02') for x in range(count)), mode)
            do_thing(handler, k, get_time(base + '00'))

        else:
            with open(image, mode) as f:
                do_thing(f, k, get_time(image))



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.
import os
from datetime import datetime
from sys import platform
import tkinter as tk
import tkinter.ttk as ttk
import tkinter.filedialog as fd
import tkinter.messagebox as mb
from typing import TYPE_CHECKING

import mountinfo
from .opendir import open_directory
from .optionsframes import CheckbuttonContainer, RadiobuttonContainer
from .outputviewer import OutputViewer
from .setupwizard import *

if TYPE_CHECKING:
    from typing import List, Optional, Type
    from . import NinfsGUI

# importing this from __init__ would cause a circular import, and it's just easier to copy this again
is_windows = platform == 'win32'

wizard_bases = {
    'cci': CCISetup,
    'cdn': CDNSetup,
    'cia': CIASetup,
    'exefs': ExeFSSetup,
    'nandctr': CTRNandImageSetup,
    'nandhac': HACNandImageSetup,
    'nandtwl': TWLNandImageSetup,
    'nandbb': BBNandImageSetup,
    'ncch': NCCHSetup,
    'romfs': RomFSSetup,
    'sd': SDFilesystemSetup,
    'sdtitle': SDTitleSetup,
    'srl': SRLSetup,
    'threedsx': ThreeDSXSetup,
}

if is_windows:
    from ctypes import windll
    from string import ascii_uppercase

    def get_unused_drives() -> 'List[str]':
        # https://stackoverflow.com/questions/827371/is-there-a-way-to-list-all-the-available-drive-letters-in-python
        drives = []
        bitmask = windll.kernel32.GetLogicalDrives()
        for letter in ascii_uppercase:
            if not bitmask & 1:
                drives.append(letter)
            bitmask >>= 1

        return drives


class WizardTypeSelector(WizardBase):
    def __init__(self, parent: 'tk.BaseWidget' = None, *, wizardcontainer: 'WizardContainer'):
        super().__init__(parent, wizardcontainer=wizardcontainer)

        self.current_type = ''

        container, type_selector, type_selector_var = self.make_option_menu('Select the content type:')
        type_selector_menu = type_selector['menu']

        container.pack(fill=tk.X, expand=True)

        def add_options(header, keys):
            type_selector_menu.add_command(label=header, state=tk.DISABLED)
            for k in keys:
                # This isn't very nice
                def changer(key):
                    def c():
                        item_info = mountinfo.get_type_info(key)
                        label_text = f'{item_info["name"]} ({item_info["info"]})'
                        self.wizardcontainer.set_next_enabled(True)
                        self.current_type = key
                        self.type_selector_var.set(label_text)
                    return c
                item_info = mountinfo.get_type_info(k)
                label_text = f'{item_info["name"]} ({item_info["info"]})'
                type_selector_menu.add_command(label='  ' + label_text, command=changer(k))

        add_options('Select a type', ())
        for cat, types in mountinfo.categories.items():
            add_options(cat, types)

        type_selector_var.set('Select a type')

        self.type_selector_var = type_selector_var

    def next_pressed(self):
        try:
            next_base = wizard_bases[self.current_type]
        except KeyError:
            mb.showerror('Derp', 'For some reason this type does not have a setup wizard module.\n'
                                 'Unfortunately this means you have to use it through the command line for now.')
            self.wizardcontainer.destroy()
        else:
            self.wizardcontainer.change_frame(next_base)


class WizardMountAdvancedOptions(tk.Toplevel):
    def __init__(self, parent: 'WizardContainer' = None, *, current: 'dict', show_dev_keys: bool):
        super().__init__(parent)

        self.wm_title('Advanced mount options')
        self.wm_resizable(width=tk.FALSE, height=tk.FALSE)

        self.wm_transient(parent)
        parent.parent.set_icon(self)
        self.grab_set()

        self.ok_clicked = False

        # prevent background issues on macOS and Linux
        outer_container = ttk.Frame(self)
        outer_container.pack(fill=tk.BOTH, expand=True)

        container = ttk.Frame(outer_container)
        container.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)

        if show_dev_keys:
            label_opt = ttk.Label(container, text='Options')
            label_opt.grid(row=0, column=0, padx=(0, 10), sticky=tk.NW)

            enabled = []
            if current['use_dev']:
                enabled.append('Use developer-unit keys')
            self.check_opt = CheckbuttonContainer(container, options=['Use developer-unit keys'], enabled=enabled)
            self.check_opt.grid(row=0, column=1, pady=(0, 10), sticky=tk.NW)
        else:
            self.check_opt = None

        if not is_windows:
            label_eua = ttk.Label(container, text='External user access')
            label_eua.grid(row=1, column=0, padx=(0, 8), sticky=tk.NW)

            opts_eua = [
                ("Don't allow other users", 'none'),
                ('Allow access by root (-o allow_root)', 'allow_root'),
                ('Allow access by other users (-o allow_other)', 'allow_other'),
            ]
            self.rb_container_eua = RadiobuttonContainer(container, options=opts_eua, default=current['user_access'])
            self.rb_container_eua.grid(row=1, column=1, pady=(0, 10))
        else:
            self.rb_container_eua = None

        def ok():
            self.ok_clicked = True
            self.destroy()

        ok_button = ttk.Button(container, text='OK', command=ok)
        ok_button.grid(row=2, column=0, columnspan=2)

        self.geometry("+%d+%d" % (parent.winfo_rootx() + 50, parent.winfo_rooty() + 50))

        self.wm_deiconify()

    def get_options(self):
        user_access = 'none'
        use_dev = False
        if self.rb_container_eua:
            user_access = self.rb_container_eua.get_selected()
        if self.check_opt:
            use_dev = self.check_opt.get_values()['Use developer-unit keys']
        return {'user_access': user_access, 'use_dev': use_dev}

    def wait_for_response(self):
        self.wait_window()
        return self.ok_clicked


class WizardMountPointSelector(WizardBase):

    mount_point_var: 'tk.StringVar'

    def __init__(self, parent: 'tk.BaseWidget' = None, *, wizardcontainer: 'WizardContainer', mounttype: 'str',
                 cmdargs: 'List[str]'):
        super().__init__(parent, wizardcontainer=wizardcontainer)
        self.mounttype = mounttype
        self.cmdargs = cmdargs

        self.adv_options = {'user_access': 'none', 'use_dev': False}

        # special case for nand types here, which should not be mounted to a drive letter
        if is_windows and not cmdargs[0].startswith('nand'):
            self.is_drive_letter = True

            drive_letters = [x + ':' for x in get_unused_drives()]

            container, drive_selector, drive_selector_var = self.make_option_menu('Select the drive letter to use:',
                                                                                  *drive_letters)
            container.pack(fill=tk.X, expand=True)

            self.mount_point_var = drive_selector_var

            if cmdargs[0] in mountinfo.supports_dev_keys:
                # the only advanced option for windows users is dev keys, so hide this button if that doesn't apply
                adv_options_button = ttk.Button(self, text='Advanced mount options', command=self.show_advanced_options)
                adv_options_button.pack(fill=tk.X, expand=True)

            self.wizardcontainer.set_next_enabled(True)

        else:
            self.is_drive_letter = False

            def callback(*_):
                mount_point = self.mount_point_var.get().strip()
                self.wizardcontainer.set_next_enabled(mount_point)

            labeltext = 'Select the directory to mount to:'
            container, mount_textbox, mount_textbox_var = self.make_directory_picker(labeltext, 'Select mountpoint')
            container.pack(fill=tk.X, expand=True)

            if (not is_windows) or cmdargs[0] in mountinfo.supports_dev_keys:
                # the only advanced option for windows users is dev keys, so hide this button if that doesn't apply
                adv_options_button = ttk.Button(self, text='Advanced mount options', command=self.show_advanced_options)
                adv_options_button.pack(fill=tk.X, expand=True)

            mount_textbox_var.trace_add('write', callback)

            self.mount_point_var = mount_textbox_var

    def show_advanced_options(self):
        adv_options_window = WizardMountAdvancedOptions(self.wizardcontainer, current=self.adv_options,
                                                        show_dev_keys=self.cmdargs[0] in mountinfo.supports_dev_keys)
        adv_options_window.focus_set()
        if adv_options_window.wait_for_response():
            self.adv_options.update(adv_options_window.get_options())

    def next_pressed(self):
        if is_windows and not self.is_drive_letter:
            if len(os.listdir(self.mount_point_var.get())) != 0:
                mb.showerror('ninfs', 'Directory must be empty.')
                return
        extra_args = []
        if self.adv_options['use_dev']:
            extra_args.append('--dev')
        if self.adv_options['user_access'] != 'none':
            extra_args.extend(('-o', self.adv_options['user_access']))
        self.wizardcontainer.mount(self.mounttype, self.cmdargs + extra_args, self.mount_point_var.get(),
                                   self.is_drive_letter)


class WizardMountStep(WizardBase):
    def __init__(self, parent: 'tk.BaseWidget' = None, *, wizardcontainer: 'WizardContainer', mounttype: str,
                 cmdargs: 'List[str]', mountpoint: str, is_drive_letter: bool):
        super().__init__(parent, wizardcontainer=wizardcontainer)

        self.wizardcontainer.set_cancel_enabled(False)

        self.mountpoint = mountpoint

        label = ttk.Label(self, text='Starting mount process...')
        label.pack(fill=tk.X, expand=True)

        self.wizardcontainer.parent.mount(mounttype, cmdargs, mountpoint, self.callback_success, self.callback_failed,
                                          is_drive_letter=is_drive_letter)

    def callback_success(self):
        opened = open_directory(self.mountpoint)
        self.wizardcontainer.destroy()

    def callback_failed(self, returncode: 'int', output: 'List[str]'):
        self.wizardcontainer.change_frame(WizardFailedMount, returncode=returncode, output=output, kind='mountfail')

    def next_pressed(self):
        pass


class WizardFailedMount(WizardBase):
    def __init__(self, parent: 'tk.BaseWidget' = None, *, wizardcontainer: 'WizardContainer',
                 returncode: 'Optional[int]', output: 'List[str]', kind: str):
        super().__init__(parent, wizardcontainer=wizardcontainer)

        if returncode:
            output.append('')
            output.append(f'Return code was {returncode}')

        self.returncode = returncode
        self.output = output

        self.rowconfigure(0, weight=0)
        self.rowconfigure(1, weight=1)
        self.columnconfigure(0, weight=1)

        kinds = {
            'mountfail': f'Failed to mount (return code {returncode}). Output is below:',
            'crash': f'The mount subprocess crashed (return code {returncode}). Output is below:'
        }
        kinds_header = {
            'mountfail': 'Failed to mount',
            'crash': 'Mount subprocess crashed'
        }
        self.set_header(kinds_header[kind])
        label = ttk.Label(self, text=kinds[kind])
        label.grid(row=0, column=0, sticky=tk.EW)

        viewer = OutputViewer(self, output=output)
        viewer.grid(row=1, column=0, sticky=tk.NSEW)

        self.wizardcontainer.next_button.configure(text='Save output to file')

        self.wizardcontainer.set_next_enabled(True)

    def next_pressed(self):
        time = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
        fn = fd.asksaveasfilename(parent=self, initialfile=f'ninfs-error_{time}.log')
        if fn:
            with open(fn, 'w', encoding='utf-8') as f:
                for line in self.output:
                    f.write(line + '\n')


class WizardContainer(tk.Toplevel):
    current_frame: 'WizardBase' = None

    def __init__(self, parent: 'NinfsGUI'):
        super().__init__(parent)
        self.parent = parent

        self.wm_withdraw()
        self.parent.set_icon(self)
        self.wm_transient(parent)

        container = ttk.Frame(self)
        container.pack(fill=tk.BOTH, expand=True)

        self.wm_title('ninfs - Mount content')
        self.wm_minsize(500, 350)

        container.rowconfigure(0, weight=0)
        container.rowconfigure(1, weight=1)
        container.rowconfigure(2, weight=0)
        container.columnconfigure(0, weight=1)

        self.header = ttk.Label(container, text='Mount new content', font=(None, 15, 'bold'), justify=tk.LEFT)
        self.header.grid(row=0, column=0, padx=10, pady=10, sticky=tk.W)

        self.container = ttk.Frame(container)
        self.container.grid(row=1, column=0, padx=10, sticky=tk.NSEW)

        self.footer_buttons = ttk.Frame(container)
        self.footer_buttons.grid(row=2, column=0, padx=10, pady=10, sticky=tk.E)

        def next_pressed():
            if self.current_frame:
                self.current_frame.next_pressed()

        self.next_button = ttk.Button(self.footer_buttons, text='Next', command=next_pressed)
        self.next_button.pack(side=tk.RIGHT)

        def cancel_pressed():
            self.destroy()

        self.cancel_button = ttk.Button(self.footer_buttons, text='Cancel', command=cancel_pressed)
        self.cancel_button.pack(side=tk.RIGHT)

        self.geometry("+%d+%d" % (parent.winfo_rootx() + 50, parent.winfo_rooty() + 50))

        self.wm_deiconify()

    def set_cancel_enabled(self, status: bool):
        self.cancel_button.configure(state=tk.NORMAL if status else tk.DISABLED)

    def set_next_enabled(self, status: bool):
        self.next_button.configure(state=tk.NORMAL if status else tk.DISABLED)

    def show_mount_point_selector(self, mounttype: 'str', cmdargs: 'List[str]'):
        self.next_button.configure(text='Mount')
        self.change_frame(WizardMountPointSelector, cmdargs=cmdargs, mounttype=mounttype)

    def mount(self, mounttype: 'str', cmdargs: 'List[str]', mountpoint: str, is_drive_letter: bool):
        self.change_frame(WizardMountStep, mounttype=mounttype, cmdargs=cmdargs, mountpoint=mountpoint,
                          is_drive_letter=is_drive_letter)

    def change_frame(self, target: 'Type[WizardBase]', *args, **kwargs):
        if self.current_frame:
            self.current_frame.pack_forget()

        self.set_cancel_enabled(True)
        self.set_next_enabled(False)
        self.current_frame = target(self.container, wizardcontainer=self, *args, **kwargs)
        self.current_frame.pack(fill=tk.BOTH, expand=True)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

"""
Mounts Executable Filesystem (ExeFS) files, creating a virtual filesystem of the ExeFS contents.
"""

import logging
from errno import ENOENT
from itertools import chain
from io import BytesIO
from stat import S_IFDIR, S_IFREG
from sys import argv
from threading import Lock
from typing import TYPE_CHECKING

import png
from pyctr.type.exefs import ExeFSReader, ExeFSFileNotFoundError, CodeDecompressionError
from pyctr.type.smdh import SMDH, InvalidSMDHError

from . import _common as _c
# _common imports these from fusepy, and prints an error if it fails; this allows less duplicated code
from ._common import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context, get_time, realpath, basename

if TYPE_CHECKING:
    from typing import BinaryIO, Dict, Union


class ExeFSMount(LoggingMixIn, Operations):
    fd = 0
    files: 'Dict[str, str]'
    special_files: 'Dict[str, Dict[str, Union[int, BinaryIO]]]'

    def __init__(self, reader: 'ExeFSReader', g_stat: dict, decompress_code: bool = False):
        self.g_stat = g_stat

        self.reader = reader
        self.decompress_code = decompress_code

        self.special_files_lock = Lock()

        # for vfs stats
        self.exefs_size = sum(x.size for x in self.reader.entries.values())

    def __del__(self, *args):
        try:
            self.reader.close()
        except AttributeError:
            pass

        with self.special_files_lock:
            for f in self.special_files.values():
                f['io'].close()

    destroy = __del__

    # TODO: maybe do this in a way that allows for multiprocessing (titledir)
    def init(self, path, data=None):
        if self.decompress_code and '.code' in self.reader.entries:
            print('ExeFS: Decompressing code...')
            try:
                res = self.reader.decompress_code()
            except CodeDecompressionError as e:
                print(f'ExeFS: Failed to decompress code: {e}')
            else:
                if res:
                    print('ExeFS: Done!')
                else:
                    print('ExeFS: No decompression needed')

        # displayed name associated with real entry name
        self.files = {'/' + x.name.replace('.', '', 1) + '.bin': x.name for x in self.reader.entries.values()}
        self.special_files = {}

        if self.reader.icon:
            icon_small = BytesIO()
            icon_large = BytesIO()

            def load_to_pypng(array, w, h):
                return png.from_array((chain.from_iterable(x) for x in array), 'RGB', {'width': w, 'height': h})

            load_to_pypng(self.reader.icon.icon_small_array, 24, 24).write(icon_small)
            load_to_pypng(self.reader.icon.icon_large_array, 48, 48).write(icon_large)

            icon_small_size = icon_small.seek(0, 2)
            icon_large_size = icon_large.seek(0, 2)

            # these names are too long to be in the exefs, so special checks can be added for them
            self.special_files['/icon_small.png'] = {'size': icon_small_size, 'io': icon_small}
            self.special_files['/icon_large.png'] = {'size': icon_large_size, 'io': icon_large}

    @_c.ensure_lower_path
    def getattr(self, path, fh=None):
        uid, gid, pid = fuse_get_context()
        if path == '/':
            st = {'st_mode': (S_IFDIR | 0o777), 'st_nlink': 2}
        else:
            if path in self.files:
                item = self.reader.entries[self.files[path]]
                size = item.size
            elif path in self.special_files:
                item = self.special_files[path]
                size = item['size']
            else:
                raise FuseOSError(ENOENT)
            st = {'st_mode': (S_IFREG | 0o666), 'st_size': size, 'st_nlink': 1}
        return {**st, **self.g_stat, 'st_uid': uid, 'st_gid': gid}

    def open(self, path, flags):
        self.fd += 1
        return self.fd

    @_c.ensure_lower_path
    def readdir(self, path, fh):
        yield from ('.', '..')
        yield from (x[1:] for x in self.files)
        yield from (x[1:] for x in self.special_files)

    @_c.ensure_lower_path
    def read(self, path, size, offset, fh):
        if path in self.files:
            with self.reader.open(self.files[path]) as f:
                f.seek(offset)
                return f.read(size)
        elif path in self.special_files:
            with self.special_files_lock:
                f = self.special_files[path]['io']
                f.seek(offset)
                return f.read(size)
        else:
            raise FuseOSError(ENOENT)

    @_c.ensure_lower_path
    def statfs(self, path):
        return {'f_bsize': 4096, 'f_frsize': 4096, 'f_blocks': self.exefs_size // 4096, 'f_bavail': 0, 'f_bfree': 0,
                'f_files': len(self.reader) + len(self.special_files)}


def main(prog: str = None, args: list = None):
    from argparse import ArgumentParser
    if args is None:
        args = argv[1:]
    parser = ArgumentParser(prog=prog, description='Mount Nintendo 3DS Executable Filesystem (ExeFS) files.',
                            parents=(_c.default_argp, _c.main_args('exefs', 'ExeFS file')))
    parser.add_argument('--decompress-code', help='decompress the .code section', action='store_true')

    a = parser.parse_args(args)
    opts = dict(_c.parse_fuse_opts(a.o))

    if a.do:
        logging.basicConfig(level=logging.DEBUG, filename=a.do)

    exefs_stat = get_time(a.exefs)

    with ExeFSReader(a.exefs) as r:
        mount = ExeFSMount(reader=r, g_stat=exefs_stat, decompress_code=a.decompress_code)
        if _c.macos or _c.windows:
            opts['fstypename'] = 'ExeFS'
            if _c.macos:
                opts['volname'] = f'Nintendo 3DS ExeFS ({basename(a.exefs)})'
            elif _c.windows:
                # volume label can only be up to 32 chars
                opts['volname'] = 'Nintendo 3DS ExeFS'
        FUSE(mount, a.mount_point, foreground=a.fg or a.d, ro=True, nothreads=True, debug=a.d,
             fsname=realpath(a.exefs).replace(',', '_'), **opts)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import tkinter as tk
import tkinter.ttk as ttk
import tkinter.filedialog as fd
from typing import TYPE_CHECKING

from ..optionsframes import CheckbuttonContainer

if TYPE_CHECKING:
    from typing import List, Tuple
    from ..wizardcontainer import WizardContainer


class WizardBase(ttk.Frame):
    def __init__(self, parent: 'tk.BaseWidget' = None, *args, wizardcontainer: 'WizardContainer', **kwargs):
        super().__init__(parent)
        self.parent = parent
        self.wizardcontainer = wizardcontainer

    def next_pressed(self):
        print('This should not be seen')

    def set_header(self, text: str):
        self.wizardcontainer.header.configure(text=text)

    def set_header_suffix(self, text: str):
        self.set_header('Mount new content - ' + text)

    def make_container(self, labeltext: str) -> 'ttk.Frame':
        container = ttk.Frame(self)

        container.rowconfigure(0, weight=1)
        container.rowconfigure(1, weight=1)
        container.columnconfigure(0, weight=1)
        container.columnconfigure(1, weight=0)

        label = ttk.Label(container, text=labeltext, justify=tk.LEFT)
        label.grid(row=0, column=0, columnspan=2, pady=(4, 0), sticky=tk.EW)

        return container

    def make_entry(self, labeltext: str, default: str = None) -> 'Tuple[ttk.Frame, ttk.Entry, tk.StringVar]':
        container = self.make_container(labeltext)

        textbox_var = tk.StringVar(self)

        textbox = ttk.Entry(container, textvariable=textbox_var)
        textbox.grid(row=1, column=0, columnspan=2, pady=4, sticky=tk.EW)

        if default:
            textbox_var.set(default)

        return container, textbox, textbox_var

    def make_file_picker(self, labeltext: str, fd_title: str,
                         default: str = None) -> 'Tuple[ttk.Frame, ttk.Entry, tk.StringVar]':
        container = self.make_container(labeltext)

        textbox_var = tk.StringVar(self)

        textbox = ttk.Entry(container, textvariable=textbox_var)
        textbox.grid(row=1, column=0, pady=4, sticky=tk.EW)

        if default:
            textbox_var.set(default)

        def choose_file():
            f = fd.askopenfilename(parent=self.parent, title=fd_title)
            if f:
                textbox_var.set(f)

        filepicker_button = ttk.Button(container, text='...', command=choose_file)
        filepicker_button.grid(row=1, column=1, pady=4)

        return container, textbox, textbox_var

    def make_directory_picker(self, labeltext: str, fd_title: str,
                              default: str = None) -> 'Tuple[ttk.Frame, ttk.Entry, tk.StringVar]':
        container = self.make_container(labeltext)

        textbox_var = tk.StringVar(self)

        textbox = ttk.Entry(container, textvariable=textbox_var)
        textbox.grid(row=1, column=0, pady=4, sticky=tk.EW)

        if default:
            textbox_var.set(default)

        def choose_file():
            f = fd.askdirectory(parent=self.parent, title=fd_title)
            if f:
                textbox_var.set(f)

        directorypicker_button = ttk.Button(container, text='...', command=choose_file)
        directorypicker_button.grid(row=1, column=1, pady=4)

        return container, textbox, textbox_var

    def make_option_menu(self, labeltext: str, *options) -> 'Tuple[ttk.Frame, ttk.OptionMenu, tk.StringVar]':
        container = self.make_container(labeltext)

        optionmenu_variable = tk.StringVar(self)

        default = None
        if options:
            default = options[0]

        optionmenu = ttk.OptionMenu(container, optionmenu_variable, default, *options)
        optionmenu.grid(row=1, column=0, columnspan=2, pady=4, sticky=tk.EW)

        return container, optionmenu, optionmenu_variable

    def make_checkbox_options(self, labeltext: str, options: 'List[str]'):
        container = ttk.Frame(self)
        label = ttk.Label(container, text=labeltext)
        label.grid(row=0, column=0, padx=(0, 4), sticky=tk.NW)

        cb = CheckbuttonContainer(container, options=options)
        cb.grid(row=0, column=1, sticky=tk.W)

        return container, cb



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

from subprocess import check_call, CalledProcessError
from sys import platform

is_windows = platform == 'win32'
is_mac = platform == 'darwin'

if is_windows:
    from os import startfile


def open_directory(path: str):
    if is_windows:
        startfile(path)
    elif is_mac:
        try:
            check_call(['/usr/bin/open', '-a', 'Finder', path])
        except CalledProcessError as e:
            return e.returncode
    else:
        # assuming linux for this, feel free to add an exception if another OS has a different method
        try:
            check_call(['xdg-open', path])
        except CalledProcessError as e:
            return e.returncode



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import tkinter as tk
import tkinter.ttk as ttk
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Dict, List, Tuple


class CheckbuttonContainer(ttk.Frame):
    def __init__(self, parent: 'tk.BaseWidget' = None, *, options: 'List[str]', enabled: 'List[str]' = None):
        super().__init__(parent)
        if not enabled:
            enabled = []

        self.variables = {}
        for opt in options:
            var = tk.BooleanVar(self)
            cb = ttk.Checkbutton(self, variable=var, text=opt)
            cb.pack(side=tk.LEFT)
            if opt in enabled:
                var.set(True)

            self.variables[opt] = var

    def get_values(self) -> 'Dict[str, bool]':
        return {x: y.get() for x, y in self.variables.items()}


class RadiobuttonContainer(ttk.Frame):
    def __init__(self, parent: 'tk.BaseWidget' = None, *, options: 'List[Tuple[str, str]]',
                 default: 'str'):
        super().__init__(parent)

        self.variable = tk.StringVar(self)
        for idx, opt in enumerate(options):
            rb = ttk.Radiobutton(self, variable=self.variable, text=opt[0], value=opt[1])
            rb.grid(row=idx, column=0, sticky=tk.W)

        self.variable.set(default)

    def get_selected(self):
        return self.variable.get()



#!/usr/bin/env python3

# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

from importlib import import_module
from inspect import cleandoc
from os import environ, makedirs
from os.path import basename, dirname, expanduser, join as pjoin, realpath
from sys import exit, argv, path, platform, hexversion, version_info, executable

_path = dirname(realpath(__file__))
if _path not in path:
    path.insert(0, _path)

import mountinfo

windows = platform in {'win32', 'cygwin'}
macos = platform == 'darwin'

python_cmd = 'py -3' if windows else 'python3'

if hexversion < 0x030800F0:
    exit('Python {0[0]}.{0[1]}.{0[2]} is not supported. Please use Python 3.8.0 or later.'.format(version_info))


def print_version():
    from __init__ import __version__
    pyver = '{0[0]}.{0[1]}.{0[2]}'.format(version_info)
    if version_info[3] != 'final':
        pyver += '{0[3][0]}{0[4]}'.format(version_info)
    # this should stay as str.format, so it runs on older versions
    print('ninfs v{0} on Python {1} - https://github.com/ihaveamac/ninfs'.format(__version__, pyver))


def exit_print_types():
    print('Please provide a mount type as the first argument.')
    print('Available mount types:')
    print()
    for cat, items in mountinfo.categories.items():
        print(cat)
        for item in items:
            info = mountinfo.get_type_info(item)
            print(f' - {item}: {info["name"]} ({info["info"]})')
    print()
    print('Additional options:')
    print('  --version                          print version')
    print('  --install-desktop-entry [PREFIX]   create desktop entry (for Linux)')
    exit(1)


def mount(mount_type: str, return_doc: bool = False) -> int:
    if mount_type in {'gui', 'gui_i_want_to_be_an_admin_pls'}:
        from gui import start_gui
        return start_gui()

    if mount_type in {'-v', '--version'}:
        # this kinda feels wrong...
        print_version()
        return 0

    # noinspection PyProtectedMember
    from pyctr.crypto import BootromNotFoundError

    if windows:
        from ctypes import windll, get_last_error

        if windll.shell32.IsUserAnAdmin():
            print('- Note: This should *not* be run as an administrator.',
                  '- The mount will not be normally accessible.',
                  '- This should be run from a non-administrator command prompt or PowerShell prompt.', sep='\n')

        # this allows for the gui parent process to send signal.CTRL_BREAK_EVENT and for this process to receive it
        try:
            import os
            parent_pid = int(environ['NINFS_GUI_PARENT_PID'])
            if windll.kernel32.AttachConsole(parent_pid) == 0:  # ATTACH_PARENT_PROCESS
                print(f'Failed to do AttachConsole({parent_pid}):', get_last_error())
                print("(Note: this most likely isn't the cause of any other issues you might have!)")
        except KeyError:
            pass

    if mount_type not in mountinfo.types and mount_type not in mountinfo.aliases:
        print_version()
        exit_print_types()

    module = import_module('mount.' + mountinfo.aliases.get(mount_type, mount_type))
    if return_doc:
        # noinspection PyTypeChecker
        return module.__doc__

    prog = None
    if __name__ != '__main__':
        prog = 'mount_' + mountinfo.aliases.get(mount_type, mount_type)
    try:
        # noinspection PyUnresolvedReferences
        return module.main(prog=prog)
    except BootromNotFoundError as e:
        print('Bootrom could not be found.',
              'Please read the README of the repository for more details.',
              'Paths checked:',
              *(' - {}'.format(x) for x in e.args[0]), sep='\n')
        return 1
    except RuntimeError as e:
        if e.args == (1,):
            pass  # assuming failed to mount and the reason would be displayed in the terminal


def create_desktop_entry(prefix: str = None):
    if windows or macos:
        print('This command is not supported for Windows or macOS.')
        return
    desktop_file = cleandoc(f'''
    [Desktop Entry]
    Name=ninfs
    Comment=Mount Nintendo contents
    Exec="{executable}" -mninfs gui
    Terminal=true
    Type=Application
    Icon=ninfs
    Categories=Utility;
    ''')
    if not prefix:
        home = expanduser('~')
        prefix = environ.get('XDG_DATA_HOME', pjoin(home, '.local', 'share'))

    app_dir = pjoin(prefix, 'applications')
    makedirs(app_dir, exist_ok=True)

    with open(pjoin(app_dir, 'ninfs.desktop'), 'w', encoding='utf-8') as o:
        print('Writing', o.name)
        o.write(desktop_file)

    for s in ('1024x1024', '128x128', '64x64', '32x32', '16x16'):
        img_dir = pjoin(prefix, 'icons', 'hicolor', s, 'apps')
        makedirs(img_dir, exist_ok=True)
        with open(pjoin(dirname(__file__), 'gui/data', s + '.png'), 'rb') as i, \
                open(pjoin(img_dir, 'ninfs.png'), 'wb') as o:
            print('Writing', o.name)
            o.write(i.read())


def main():
    exit(mount(basename(argv[0])[6:].lower()))


def gui(_allow_admin: bool = False):
    if len(argv) < 2 or argv[1] in {'gui', 'gui_i_want_to_be_an_admin_pls'}:
        from gui import start_gui
        exit(start_gui())
    else:
        exit(mount(argv.pop(1).lower()))


if __name__ == '__main__':
    # path fun times

    if len(argv) < 2:
        exit_print_types()

    exit(mount(argv.pop(1).lower()))



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import logging
import time
from argparse import ArgumentParser, SUPPRESS
from errno import EROFS
from functools import wraps
from io import BufferedIOBase
from os import stat, stat_result
from os.path import basename, realpath as real_realpath
from sys import exit, platform
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from os import PathLike
    from typing import BinaryIO, Generator, Tuple, Union
    # this is a lazy way to make type checkers stop complaining
    BufferedIOBase = BinaryIO

windows = platform in {'win32', 'cygwin'}
macos = platform == 'darwin'

python_cmd = 'py -3' if windows else 'python3'

# TODO: switch to use_ns in all scripts
# noinspection PyBroadException
try:
    from fuse import FUSE, FuseOSError, Operations, fuse_get_context
except Exception as e:
    exit(f'Failed to import the fuse module:\n'
         f'{type(e).__name__}: {e}')


def realpath(path):
    try:
        return real_realpath(path)
    except OSError:
        # can happen on Windows when using it on files inside a WinFsp mount
        pass
    return path


def get_time(path: 'Union[str, PathLike, stat_result]'):
    try:
        if not isinstance(path, stat_result):
            res = stat(path)
        else:
            res = path
        return {'st_ctime': int(res.st_ctime), 'st_mtime': int(res.st_mtime), 'st_atime': int(res.st_atime)}
    except OSError:
        # sometimes os.stat can't be used with a path, such as Windows physical drives
        #   so we need to fake the result
        now = int(time.time())
        return {'st_ctime': now, 'st_mtime': now, 'st_atime': now}


# custom LoggingMixIn modified from the original fusepy, to suppress certain entries.
class LoggingMixIn:
    log = logging.getLogger('fuse.log-mixin')

    def __call__(self, op, path, *args):
        if op != 'access':
            self.log.debug('-> %s %s %s', op, path, repr(args))
        ret = '[Unhandled Exception]'
        try:
            ret = getattr(self, op)(path, *args)
            return ret
        except OSError as e:
            ret = str(e)
            raise
        finally:
            if op != 'access':
                self.log.debug('<- %s %s', op, repr(ret))


default_argp = ArgumentParser(add_help=False)
default_argp.add_argument('-f', '--fg', help='run in foreground', action='store_true')
default_argp.add_argument('-d', help='debug output (fuse/winfsp log)', action='store_true')
default_argp.add_argument('--do', help=SUPPRESS, default=None)  # debugging using python logging
default_argp.add_argument('-o', metavar='OPTIONS', help='mount options')

readonly_argp = ArgumentParser(add_help=False)
readonly_argp.add_argument('-r', '--ro', help='mount read-only', action='store_true')

ctrcrypto_argp = ArgumentParser(add_help=False)
ctrcrypto_argp.add_argument('--boot9', help='path to boot9.bin')
ctrcrypto_argp.add_argument('--dev', help='use dev keys', action='store_const', const=1, default=0)

seeddb_argp = ArgumentParser(add_help=False)
seeddb_argp_group = seeddb_argp.add_mutually_exclusive_group()
seeddb_argp_group.add_argument('--seeddb', help='path to seeddb.bin')
seeddb_argp_group.add_argument('--seed', help='seed as hexstring')


def main_args(name: str, help: str) -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(name, help=help)
    parser.add_argument('mount_point', help='mount point')
    return parser


def load_custom_boot9(path: str, dev: bool = False):
    """Load keys from a custom ARM9 bootROM path."""
    if path:
        from pyctr.crypto import CryptoEngine
        # doing this will set up the keys for all future CryptoEngine objects
        CryptoEngine(boot9=path, dev=dev)


# aren't type hints great?
def parse_fuse_opts(opts) -> 'Generator[Tuple[str, Union[str, bool]], None, None]':
    if not opts:
        return
    for arg in opts.split(','):
        if arg:  # leaves out empty ones
            separated = arg.split('=', maxsplit=1)
            yield separated[0], True if len(separated) == 1 else separated[1]


def remove_first_dir(path: str) -> str:
    sep = path.find('/', 1)
    if sep == -1:
        return '/'
    else:
        return path[sep:]


def get_first_dir(path: str) -> str:
    sep = path.find('/', 1)
    if sep == -1:
        return path
    else:
        return path[:sep]


def ensure_lower_path(method):
    @wraps(method)
    def wrapper(self, path, *args, **kwargs):
        return method(self, path.lower(), *args, **kwargs)
    return wrapper


def raise_on_readonly(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if self.readonly:
            raise FuseOSError(EROFS)
        return method(self, *args, **kwargs)
    return wrapper


def _raise_if_closed(method):
    @wraps(method)
    def decorator(self, *args, **kwargs):
        if self.closed:
            raise ValueError('I/O operation on closed file.')
        return method(self, *args, **kwargs)
    return decorator


class VirtualFileWrapper(BufferedIOBase):
    """Wrapper for a FUSE Operations class for things that need a file-like object."""

    _seek = 0

    # noinspection PyMissingConstructor
    def __init__(self, fuse_op: Operations, path: str, size: int):
        self.fuse_op = fuse_op
        self.path = path
        self.size = size

    @_raise_if_closed
    def read(self, size: int = -1) -> bytes:
        if size == -1:
            size = self.size - self._seek
        data = self.fuse_op.read(self.path, size, self._seek, 0)
        self._seek += len(data)
        return data

    read1 = read  # probably make this act like read1 should, but this for now enables some other things to work

    @_raise_if_closed
    def seek(self, seek: int, whence: int = 0) -> int:
        if whence == 0:
            if seek < 0:
                raise ValueError(f'negative seek value {seek}')
            self._seek = min(seek, self.size)
        elif whence == 1:
            self._seek = max(self._seek + seek, 0)
        elif whence == 2:
            self._seek = max(self.size + seek, 0)
        return self._seek

    @_raise_if_closed
    def tell(self) -> int:
        return self._seek

    @_raise_if_closed
    def readable(self) -> bool:
        return True

    @_raise_if_closed
    def writable(self) -> bool:
        try:
            # types that support writing will have this attribute
            return self.fuse_op.readonly
        except AttributeError:
            return False

    @_raise_if_closed
    def seekable(self) -> bool:
        return True


class SplitFileHandler(BufferedIOBase):
    _fake_seek = 0
    _seek_info = (0, 0)

    def __init__(self, names, mode='rb'):
        self.mode = mode
        self._files = []
        curr_offset = 0
        self._names = tuple(names)
        for idx, f in enumerate(self._names):
            s = stat(f)
            self._files.append((idx, curr_offset, s.st_size))
            curr_offset += s.st_size
        self._total_size = curr_offset

    def _calc_seek(self, pos):
        for idx, info in enumerate(self._files):
            if info[1] <= pos < info[1] + info[2]:
                self._fake_seek = pos
                self._seek_info = (idx, pos - info[1])
                break

    def seek(self, pos, whence=0):
        if whence == 0:
            if pos < 0:
                raise ValueError('negative seek value')
            self._calc_seek(pos)
        elif whence == 1:
            if self._fake_seek - pos < 0:
                pos = 0
            self._calc_seek(self._fake_seek + pos)
        elif whence == 2:
            if self._total_size + pos < 0:
                pos = -self._total_size
            self._calc_seek(self._total_size + pos)
        else:
            if isinstance(whence, int):
                raise ValueError(f'whence value {whence} unsupported')
            else:
                raise TypeError(f'an integer is required (got type {type(whence).__name__})')
        return self._fake_seek

    @_raise_if_closed
    def tell(self):
        return self._fake_seek

    def read(self, count=-1):
        if count == -1:
            count = self._total_size - count
        if self._fake_seek + count > self._total_size:
            count = self._total_size - self._fake_seek

        left = count
        curr = self._seek_info

        full_data = []

        while left:
            info = self._files[curr[0]]
            real_seek = self._fake_seek - info[1]
            to_read = min(info[2] - real_seek, left)

            with open(self._names[curr[0]], 'rb') as f:
                f.seek(real_seek)
                full_data.append(f.read(to_read))

            self._fake_seek += to_read
            try:
                curr = self._files[curr[0] + 1]
                left -= to_read
            except IndexError:
                break  # EOF

        # TODO: make this more efficient
        self._calc_seek(self._fake_seek)

        return b''.join(full_data)

    def write(self, data: bytes):
        left = len(data)
        total = left
        curr = self._seek_info

        while left:
            info = self._files[curr[0]]
            real_seek = self._fake_seek - info[1]
            to_write = min(info[2] - real_seek, left)

            with open(self._names[curr[0]], 'rb+') as f:
                f.seek(real_seek)
                f.write(data[total - left:total - left + to_write])

            self._fake_seek += to_write
            try:
                curr = self._files[curr[0] + 1]
                left -= to_write
            except IndexError:
                break  # EOF

        # TODO: make this more efficient
        self._calc_seek(self._fake_seek)

        return total - left

    @_raise_if_closed
    def readable(self) -> bool:
        return 'r' in self.mode

    def writable(self) -> bool:
        return 'w' in self.mode or 'a' in self.mode

    @_raise_if_closed
    def seekable(self) -> bool:
        return True


class RawDeviceHandler(BufferedIOBase):
    """(NYI) Handler for easier IO access with raw devices by aligning reads and writes to the sector size."""

    _seek = 0

    def __init__(self, fh: 'BinaryIO', sector_size: int = 0x200):
        self._fh = fh
        self._sector_size = sector_size

    @_raise_if_closed
    def seek(self, seek: int, whence: int = 0) -> int:
        if whence == 0:
            if seek < 0:
                raise ValueError(f'negative seek value {seek}')
            self._seek = seek
        elif whence == 1:
            self._seek = max(self._seek + seek, 0)
        elif whence == 2:
            # this doesn't work...
            # maybe if the size is known, this could be based off that instead?
            raise NotImplementedError("can't seek from the ending on a block device")
        return self._seek

    @_raise_if_closed
    def tell(self) -> int:
        return self._seek

    @_raise_if_closed
    def readable(self) -> bool:
        return self._fh.readable()

    @_raise_if_closed
    def writable(self) -> bool:
        return self._fh.writable()

    @_raise_if_closed
    def seekable(self) -> bool:
        return self._fh.seekable()

    @_raise_if_closed
    def read(self, size: int = -1) -> bytes:
        if size == -1:
            raise NotImplementedError("can't read without a specified size")

        before = self._seek % self._sector_size
        after = 0
        total = before + size
        if total % self._sector_size:
            after = self._sector_size - (total % self._sector_size)
            total += after

        self._seek += size



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import json
import tkinter as tk
import tkinter.ttk as ttk
import webbrowser
from typing import TYPE_CHECKING
from urllib.request import urlopen, Request

from pkg_resources import parse_version

from .outputviewer import OutputViewer
from .setupwizard import WizardBase
from .wizardcontainer import WizardContainer

if TYPE_CHECKING:
    from typing import Tuple
    # noinspection PyProtectedMember
    from pkg_resources._vendor.packaging.version import Version
    from http.client import HTTPResponse

    from . import NinfsGUI

from __init__ import __version__

version: 'Version' = parse_version(__version__)


class UpdateNotificationWindow(WizardBase):
    def __init__(self, parent: 'tk.BaseWidget' = None, *, wizardcontainer: 'WizardContainer', releaseinfo: 'dict'):
        super().__init__(parent, wizardcontainer=wizardcontainer)

        self.releaseinfo = releaseinfo

        self.set_header('New update available - ' + releaseinfo['tag_name'])

        self.rowconfigure(0, weight=0)
        self.rowconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)

        label = ttk.Label(self, text='A new update for ninfs is available!')
        label.grid(row=0, column=0, sticky=tk.EW)

        rel_body_full = releaseinfo['body']
        rel_body = rel_body_full[:rel_body_full.find('------')].replace('\r\n', '\n').strip()

        viewer = OutputViewer(self, output=[rel_body])
        viewer.grid(row=1, column=0, sticky=tk.NSEW)

        self.wizardcontainer.next_button.configure(text='Open release info')
        self.wizardcontainer.set_next_enabled(True)

    def next_pressed(self):
        webbrowser.open(self.releaseinfo['html_url'])


def get_latest_release() -> 'dict':
    url = 'https://api.github.com/repos/ihaveamac/ninfs/releases'
    if not version.is_prerelease:
        url += '/latest'

    print('UPDATE: Requesting', url)
    req = Request(url, headers={'Accept': 'application/vnd.github.v3+json'})
    with urlopen(req) as u:  # type: HTTPResponse
        data = json.load(u)

    if version.is_prerelease:
        data = data[0]

    return data


def thread_update_check(gui: 'NinfsGUI'):
    rel = get_latest_release()
    latest_version: 'Version' = parse_version(rel['tag_name'])

    print('UPDATE: Latest version:', latest_version)
    if latest_version > version:
        wizard_window = WizardContainer(gui)
        wizard_window.change_frame(UpdateNotificationWindow, releaseinfo=rel)
        wizard_window.focus()



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import tkinter as tk
from typing import TYPE_CHECKING

from .base import WizardBase
from .. import supportfiles

if TYPE_CHECKING:
    from .. import WizardContainer


class NCCHSetup(WizardBase):
    def __init__(self, parent: 'tk.BaseWidget' = None, *, wizardcontainer: 'WizardContainer'):
        super().__init__(parent, wizardcontainer=wizardcontainer)

        def callback(*_):
            main_file = self.main_textbox_var.get().strip()
            b9_file = self.b9_textbox_var.get().strip()
            self.wizardcontainer.set_next_enabled(main_file and b9_file)

        main_container, main_textbox, main_textbox_var = self.make_file_picker('Select the NCCH file:',
                                                                               'Select NCCH file')
        main_container.pack(fill=tk.X, expand=True)

        b9_container, b9_textbox, b9_textbox_var = self.make_file_picker('Select the boot9 file:', 'Select boot9 file')
        b9_container.pack(fill=tk.X, expand=True)

        seeddb_container, seeddb_textbox, seeddb_textbox_var = self.make_file_picker('Select the seeddb file:',
                                                                                     'Select seeddb file')
        seeddb_container.pack(fill=tk.X, expand=True)

        seed_container, seed_textbox, seed_textbox_var = self.make_entry('OR Enter the seed '
                                                                         '(optional if seeddb is used):')
        seed_container.pack(fill=tk.X, expand=True)

        self.main_textbox_var = main_textbox_var
        self.b9_textbox_var = b9_textbox_var
        self.seeddb_textbox_var = seeddb_textbox_var
        self.seed_textbox_var = seed_textbox_var

        main_textbox_var.trace_add('write', callback)
        b9_textbox_var.trace_add('write', callback)

        b9_textbox_var.set(supportfiles.last_b9_file)
        seeddb_textbox_var.set(supportfiles.last_seeddb_file)

        self.set_header_suffix('NCCH')

    def next_pressed(self):
        main_file = self.main_textbox_var.get().strip()
        b9_file = self.b9_textbox_var.get().strip()
        seeddb_file = self.seeddb_textbox_var.get().strip()
        seed = self.seed_textbox_var.get().replace(' ', '')

        if b9_file:
            supportfiles.last_b9_file = b9_file
        if seeddb_file:
            supportfiles.last_seeddb_file = seeddb_file

        args = ['ncch', main_file]
        if b9_file:
            args += ['--boot9', b9_file]
        if seed:
            args += ['--seed', seed]
        elif seeddb_file:
            args += ['--seeddb', seeddb_file]

        self.wizardcontainer.show_mount_point_selector('NCCH', args)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import sys
import tkinter as tk
import tkinter.ttk as ttk
import webbrowser
from os.path import join
from importlib.metadata import metadata
from typing import TYPE_CHECKING

from .osver import get_os_ver
# "from .. import" didn't work :/
from __init__ import __copyright__ as ninfs_copyright
from __init__ import __version__ as ninfs_version

frozen_using = None
if getattr(sys, 'frozen', False):
    if getattr(sys, '_MEIPASS', False):
        frozen_using = 'PyInstaller'
    else:
        # cx-Freeze probably
        frozen_using = 'cx-Freeze'

if TYPE_CHECKING:
    from . import NinfsGUI

pad = 10

python_version = sys.version.split()[0]
pybits = 64 if sys.maxsize > 0xFFFFFFFF else 32
os_ver = get_os_ver()


class LicenseViewer(ttk.Frame):
    def __init__(self, parent: 'tk.BaseWidget' = None, *, text: str):
        super().__init__(parent)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=0)

        scrollbar = ttk.Scrollbar(self, orient=tk.VERTICAL)
        scrollbar.grid(row=0, column=1, sticky=tk.NSEW)

        textarea = tk.Text(self, wrap='word', yscrollcommand=scrollbar.set)
        textarea.grid(row=0, column=0, sticky=tk.NSEW)

        scrollbar.configure(command=textarea.yview)

        textarea.insert(tk.END, text)

        textarea.configure(state=tk.DISABLED)


class NinfsAbout(tk.Toplevel):
    def __init__(self, parent: 'NinfsGUI' = None):
        super().__init__(parent)
        self.parent = parent

        self.wm_withdraw()
        self.parent.set_icon(self)
        self.wm_transient(self.parent)

        container = ttk.Frame(self)
        container.pack(fill=tk.BOTH, expand=True)

        self.wm_title('About ninfs')
        self.wm_resizable(width=tk.FALSE, height=tk.FALSE)

        header_label = ttk.Label(container, text=f'ninfs {ninfs_version}', font=(None, 15, 'bold'))
        header_label.grid(row=0, column=0, padx=pad, pady=pad, sticky=tk.W)

        running_on = f'Running on Python {python_version} {pybits}-bit'
        if frozen_using:
            running_on += f' (frozen using {frozen_using})'
        version_label = ttk.Label(container, text=running_on)
        version_label.grid(row=1, column=0, padx=pad, pady=(0, pad), sticky=tk.W)

        copyright_label = ttk.Label(container, text='This program uses several libraries and modules, which have '
                                                    'their licenses below.')
        copyright_label.grid(row=2, column=0, padx=pad, pady=(0, pad), sticky=tk.W)

        # tab name, license file name, url
        info = [
            (f'ninfs', ninfs_version, 'ninfs.md', 'https://github.com/ihaveamac/ninfs',
             'ninfs - Copyright (c) 2017-2021 Ian Burgwin'),
            (f'WinFsp', '2023', 'winfsp.txt', 'https://github.com/billziss-gh/winfsp',
             'WinFsp - Windows File System Proxy, Copyright (C) Bill Zissimopoulos'),
            (f'pycryptodomex', metadata('pycryptodomex')['Version'], 'pycryptodome.rst',
             'https://github.com/Legrandin/pycryptodome', 'PyCryptodome - multiple licenses'),
            (f'pyctr', metadata('pyctr')['Version'], 'pyctr', 'https://github.com/ihaveamac/pyctr',
             'pyctr - Copyright (c) 2017-2021 Ian Burgwin'),
            ('haccrypto', metadata('haccrypto')['Version'], 'haccrypto.md', 'https://github.com/luigoalma/haccrypto',
             'haccrypto - Copyright (c) 2017-2021 Ian Burgwin & Copyright (c) 2020-2021 Luis Marques')
        ]

        license_notebook = ttk.Notebook(container)
        license_notebook.grid(row=3, column=0, padx=pad, pady=(0, pad))

        def cmd_maker(do_url):
            def func():
                webbrowser.open(do_url)
            return func

        for proj_name, proj_version, license_file, url, header in info:
            frame = ttk.Frame(license_notebook)
            license_notebook.add(frame, text=proj_name + ' ' + proj_version)

            license_header_label = ttk.Label(frame, text=header)
            license_header_label.grid(row=0, sticky=tk.W, padx=pad//2, pady=pad//2)

            url_button = ttk.Button(frame,
                                    text='Open website - ' + url,
                                    command=cmd_maker(url))
            url_button.grid(row=1)

            with open(parent.get_data_file(join('licenses', license_file)), 'r', encoding='utf-8') as f:
                license_frame = LicenseViewer(frame, text=f.read())
                license_frame.grid(row=2)

        self.geometry("+%d+%d" % (parent.winfo_rootx() + 50, parent.winfo_rooty() + 50))

        self.wm_deiconify()



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

from .base import WizardBase

from .cci import CCISetup
from .cdn import CDNSetup
from .cia import CIASetup
from .exefs import ExeFSSetup
from .nandctr import CTRNandImageSetup
from .nandhac import HACNandImageSetup
from .nandtwl import TWLNandImageSetup
from .nandbb import BBNandImageSetup
from .ncch import NCCHSetup
from .romfs import RomFSSetup
from .sd import SDFilesystemSetup
from .sdtitle import SDTitleSetup
from .srl import SRLSetup
from .threedsx import ThreeDSXSetup



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import sys
from os.path import dirname, join
from os import environ

if len(sys.argv) > 1:
    # Ignore `-psn_0_#######` which is added if macOS App Translocation is in effect
    if sys.argv[1].startswith('-psn'):
        del sys.argv[1]

if getattr(sys, 'frozen', False):
    if hasattr(sys, '_MEIPASS'):
        # PyInstaller
        pass
    else:
        # cx_Freeze probably
        sys.path.insert(0, join(dirname(sys.executable), 'lib', 'ninfs'))

        if sys.platform == 'win32':
            # this will try to fix loading tkinter in paths containing non-latin characters
            environ['TCL_LIBRARY'] = 'lib/tkinter/tcl8.6'

from main import gui
gui()



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

from configparser import ConfigParser
from os import environ, makedirs
from os.path import expanduser, isdir, join
from sys import platform
from threading import Lock

__all__ = ['get_bool', 'set_bool']

CONFIG_FILENAME = 'config.ini'

home = expanduser('~')

lock = Lock()

if platform == 'win32':
    config_dir = join(environ['APPDATA'], 'ninfs')
elif platform == 'darwin':
    config_dir = join(home, 'Library', 'Application Support', 'ninfs')
else:
    # probably linux or bsd or something
    # if by some chance an OS uses different paths, feel free to let me know or make a PR
    config_root = environ.get('XDG_CONFIG_HOME')
    if not config_root:
        # check other paths in XDG_CONFIG_DIRS to see if ninfs already exists in one of them
        config_roots = environ.get('XDG_CONFIG_DIRS')
        if not config_roots:
            config_roots = '/etc/xdg'
        config_paths = config_roots.split(':')
        for path in config_paths:
            d = join(path, 'ninfs')
            if isdir(d):
                config_root = d
                break
    # check again to see if it was set
    if not config_root:
        config_root = join(home, '.config')
    config_dir = join(config_root, 'ninfs')

makedirs(config_dir, exist_ok=True)

config_file = join(config_dir, CONFIG_FILENAME)

parser = ConfigParser()

# defaults
parser['update'] = {}
parser['update']['onlinecheck'] = 'false'
parser['internal'] = {}
parser['internal']['askedonlinecheck'] = 'false'


def save_config():
    with lock:
        print('Saving to:', config_file)
        with open(config_file, 'w') as f:
            parser.write(f)


def get_bool(section: 'str', key: 'str'):
    return parser.getboolean(section, key)


def set_bool(section: 'str', key: 'str', value: bool):
    parser.set(section, key, 'true' if value else 'false')
    save_config()


# load user config if possible
loaded = parser.read(config_file)
if not loaded:
    save_config()



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import platform
import sys


def get_os_ver():
    if sys.platform == 'win32':
        import winreg

        ver = platform.win32_ver()
        marketing_ver = ver[0]
        service_pack = ver[2]

        os_ver = 'Windows ' + marketing_ver
        if marketing_ver == '10':
            k = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r'SOFTWARE\Microsoft\Windows NT\CurrentVersion')
            try:
                r = winreg.QueryValueEx(k, 'ReleaseId')[0]
            except FileNotFoundError:
                r = '1507'  # assuming RTM, since this key only exists in 1511 and up
            os_ver += ', version ' + r
        else:
            if service_pack != 'SP0':
                os_ver += ' ' + service_pack
    elif sys.platform == 'darwin':
        ver = platform.mac_ver()
        os_ver = f'macOS {ver[0]} on {ver[2]}'
    else:
        ver = platform.uname()
        os_ver = f'{ver.system} {ver.release} on {ver.machine}'

    return os_ver



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

"""
Mounts installed SD title contents, creating a virtual filesystem of decrypted contents (if encrypted).
"""

import logging
from errno import ENOENT
from glob import glob
from os.path import isfile, join
from stat import S_IFDIR, S_IFREG
from sys import argv
from typing import TYPE_CHECKING

from pyctr.crypto import load_seeddb
from pyctr.type.sdtitle import SDTitleReader, SDTitleSection

from . import _common as _c
# _common imports these from fusepy, and prints an error if it fails; this allows less duplicated code
from ._common import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context, get_time, load_custom_boot9, \
    realpath
from .ncch import NCCHContainerMount
from .srl import SRLMount

if TYPE_CHECKING:
    from os import DirEntry
    from typing import Dict, Tuple, Union


class SDTitleContentsMount(LoggingMixIn, Operations):
    fd = 0
    total_size = 0

    def __init__(self, reader: 'SDTitleReader', g_stat: dict):
        self.dirs: Dict[str, Union[NCCHContainerMount, SRLMount]] = {}
        self.files: Dict[str, Tuple[Union[int, SDTitleSection], int, int]] = {}

        # get status change, modify, and file access times
        self.g_stat = g_stat

        self.reader = reader

    def __del__(self, *args):
        try:
            self.reader.close()
        except AttributeError:
            pass

    destroy = __del__

    def init(self, path):
        def add_file(name: str, section: 'Union[SDTitleSection, int]', added_offset: int = 0):
            # added offset is used for a few things like meta icon and tmdchunks
            if section >= 0:
                size = self.reader.content_info[section].size
            else:
                with self.reader.open_raw_section(section) as f:
                    size = f.seek(0, 2)
            self.files[name] = (section, added_offset, size - added_offset)

        add_file('/tmd.bin', SDTitleSection.TitleMetadata)
        add_file('/tmdchunks.bin', SDTitleSection.TitleMetadata, 0xB04)

        for record in self.reader.content_info:
            dirname = f'/{record.cindex:04x}.{record.id}'
            is_srl = record.cindex == 0 and self.reader.tmd.title_id[3:5] == '48'
            file_ext = 'nds' if is_srl else 'ncch'
            filename = f'{dirname}.{file_ext}'
            add_file(filename, record.cindex)
            try:
                if is_srl:
                    # special case for SRL contents
                    srl_fp = self.reader.open_raw_section(record.cindex)
                    self.dirs[dirname] = SRLMount(srl_fp, g_stat=self.g_stat)
                else:
                    mount = NCCHContainerMount(self.reader.contents[record.cindex], g_stat=self.g_stat)
                    mount.init(path)
                    self.dirs[dirname] = mount
            except Exception as e:
                print(f'Failed to mount {filename}: {type(e).__name__}: {e}')

            self.total_size += record.size

    @_c.ensure_lower_path
    def getattr(self, path, fh=None):
        first_dir = _c.get_first_dir(path)
        if first_dir in self.dirs:
            return self.dirs[first_dir].getattr(_c.remove_first_dir(path), fh)
        uid, gid, pid = fuse_get_context()
        if path == '/' or path in self.dirs:
            st = {'st_mode': (S_IFDIR | 0o777), 'st_nlink': 2}
        elif path in self.files:
            st = {'st_mode': (S_IFREG | 0o666), 'st_size': self.files[path][2], 'st_nlink': 1}
        else:
            raise FuseOSError(ENOENT)
        return {**st, **self.g_stat, 'st_uid': uid, 'st_gid': gid}

    def open(self, path, flags):
        self.fd += 1
        return self.fd

    @_c.ensure_lower_path
    def readdir(self, path, fh):
        first_dir = _c.get_first_dir(path)
        if first_dir in self.dirs:
            yield from self.dirs[first_dir].readdir(_c.remove_first_dir(path), fh)
        else:
            yield from ('.', '..')
            yield from (x[1:] for x in self.files)
            yield from (x[1:] for x in self.dirs)

    @_c.ensure_lower_path
    def read(self, path, size, offset, fh):
        first_dir = _c.get_first_dir(path)
        if first_dir in self.dirs:
            return self.dirs[first_dir].read(_c.remove_first_dir(path), size, offset, fh)

        section = self.files[path]
        with self.reader.open_raw_section(section[0]) as f:
            f.seek(offset + section[1])
            return f.read(size)

    @_c.ensure_lower_path
    def statfs(self, path):
        first_dir = _c.get_first_dir(path)
        if first_dir in self.dirs:
            return self.dirs[first_dir].statfs(_c.remove_first_dir(path))
        return {'f_bsize': 4096, 'f_frsize': 4096, 'f_blocks': self.total_size // 4096, 'f_bavail': 0, 'f_bfree': 0,
                'f_files': len(self.files)}


def main(prog: str = None, args: list = None):
    from argparse import ArgumentParser
    if args is None:
        args = argv[1:]
    parser = ArgumentParser(prog=prog, description='Mount Nintendo 3DS installed SD title contents.',
                            parents=(_c.default_argp, _c.ctrcrypto_argp, _c.seeddb_argp,
                                     _c.main_args('content', 'tmd file or directory with SD title contents')))

    a = parser.parse_args(args)
    opts = dict(_c.parse_fuse_opts(a.o))

    if a.do:
        logging.basicConfig(level=logging.DEBUG, filename=a.do)

    if isfile(a.content):
        tmd_file = a.content
    else:
        tmd_file = None
        tmds = glob(join(a.content, '*.tmd'))
        if tmds:
            # if there end up being multiple, this should use the first one that's found, which is probably the
            # active one used by the system right now
            tmd_file = tmds[0]
        else:
            exit(f'Could not find a tmd in {a.content}')

    sdtitle_stat = get_time(a.content)

    load_custom_boot9(a.boot9)

    if a.seeddb:
        load_seeddb(a.seeddb)

    with SDTitleReader(tmd_file, dev=a.dev, seed=a.seed, case_insensitive=True) as r:
        mount = SDTitleContentsMount(reader=r, g_stat=sdtitle_stat)
        if _c.macos or _c.windows:
            opts['fstypename'] = 'SDT'
            if _c.macos:
                display = r.tmd.title_id.upper()
                try:
                    title = r.contents[0].exefs.icon.get_app_title()
                    display += f'; ' + r.contents[0].product_code
                    if title.short_desc != 'unknown':
                        display += '; ' + title.short_desc
                except:
                    pass
                opts['volname'] = f'SD Title Contents ({display})'
            elif _c.windows:
                # volume label can only be up to 32 chars
                try:
                    title = r.contents[0].exefs.icon.get_app_title().short_desc
                    if len(title) > 21:
                        title = title[0:20] + '\u2026'  # ellipsis
                    display = title
                except:
                    display = r.tmd.title_id.upper()
                opts['volname'] = f'SD Title ({display})'
        FUSE(mount, a.mount_point, foreground=a.fg or a.d, ro=True, nothreads=True, debug=a.d,
             fsname=realpath(tmd_file).replace(',', '_'), **opts)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

from contextlib import suppress
from os import environ
from os.path import isfile, join

from pyctr.util import config_dirs

b9_paths = []
seeddb_paths = []
for p in config_dirs:
    b9_paths.append(join(p, 'boot9.bin'))
    b9_paths.append(join(p, 'boot9_prot.bin'))
    seeddb_paths.append(join(p, 'seeddb.bin'))

with suppress(KeyError):
    b9_paths.insert(0, environ['BOOT9_PATH'])

with suppress(KeyError):
    seeddb_paths.insert(0, environ['SEEDDB_PATH'])

last_b9_file = ''
last_seeddb_file = ''

for p in b9_paths:
    if isfile(p):
        last_b9_file = p
        break

for p in seeddb_paths:
    if isfile(p):
        last_seeddb_file = p
        break



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

"""
Mounts CTR Cart Image (CCI, ".3ds") files, creating a virtual filesystem of separate partitions.
"""

import logging
from errno import ENOENT
from stat import S_IFDIR, S_IFREG
from sys import argv
from typing import TYPE_CHECKING

from pyctr.type.cci import CCIReader, CCISection

from . import _common as _c
# _common imports these from fusepy, and prints an error if it fails; this allows less duplicated code
from ._common import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context, get_time, load_custom_boot9, \
    realpath
from .ncch import NCCHContainerMount

if TYPE_CHECKING:
    from typing import Dict


class CTRCartImageMount(LoggingMixIn, Operations):
    fd = 0

    def __init__(self, reader: 'CCIReader', g_stat: dict):
        self.dirs: Dict[str, NCCHContainerMount] = {}
        self.files: Dict[str, CCISection] = {}

        # get status change, modify, and file access times
        self.g_stat = g_stat

        self.reader = reader

    def __del__(self, *args):
        try:
            self.f.close()
        except AttributeError:
            pass

    destroy = __del__

    def init(self, path):
        ncsd_part_names = ('game', 'manual', 'dlp', 'unk', 'unk', 'unk', 'update_n3ds', 'update_o3ds')

        def add_file(name: str, section: 'CCISection'):
            self.files[name] = section

        add_file('/ncsd.bin', CCISection.Header)
        add_file('/cardinfo.bin', CCISection.CardInfo)
        add_file('/devinfo.bin', CCISection.DevInfo)
        for part, ncch_reader in self.reader.contents.items():
            dirname = f'/content{part}.{ncsd_part_names[part]}'
            filename = dirname + '.ncch'

            add_file(filename, part)
            try:
                mount = NCCHContainerMount(ncch_reader, g_stat=self.g_stat)
                mount.init(path)
                self.dirs[dirname] = mount
            except Exception as e:
                print(f'Failed to mount {filename}: {type(e).__name__}: {e}')

    @_c.ensure_lower_path
    def getattr(self, path, fh=None):
        first_dir = _c.get_first_dir(path)
        if first_dir in self.dirs:
            return self.dirs[first_dir].getattr(_c.remove_first_dir(path), fh)
        uid, gid, pid = fuse_get_context()
        if path == '/':
            st = {'st_mode': (S_IFDIR | 0o777), 'st_nlink': 2}
        elif path in self.files:
            st = {'st_mode': (S_IFREG | 0o666), 'st_size': self.reader.sections[self.files[path]].size, 'st_nlink': 1}
        else:
            raise FuseOSError(ENOENT)
        return {**st, **self.g_stat, 'st_uid': uid, 'st_gid': gid}

    def open(self, path, flags):
        self.fd += 1
        return self.fd

    @_c.ensure_lower_path
    def readdir(self, path, fh):
        first_dir = _c.get_first_dir(path)
        if first_dir in self.dirs:
            yield from self.dirs[first_dir].readdir(_c.remove_first_dir(path), fh)
        else:
            yield from ('.', '..')
            yield from (x[1:] for x in self.files)
            yield from (x[1:] for x in self.dirs)

    @_c.ensure_lower_path
    def read(self, path, size, offset, fh):
        first_dir = _c.get_first_dir(path)
        if first_dir in self.dirs:
            return self.dirs[first_dir].read(_c.remove_first_dir(path), size, offset, fh)

        section = self.files[path]
        with self.reader.open_raw_section(section) as f:
            f.seek(offset)
            return f.read(size)

    @_c.ensure_lower_path
    def statfs(self, path):
        first_dir = _c.get_first_dir(path)
        if first_dir in self.dirs:
            return self.dirs[first_dir].statfs(_c.remove_first_dir(path))
        return {'f_bsize': 4096, 'f_frsize': 4096, 'f_blocks': self.reader.image_size // 4096, 'f_bavail': 0,
                'f_bfree': 0, 'f_files': len(self.files)}


def main(prog: str = None, args: list = None):
    from argparse import ArgumentParser
    if args is None:
        args = argv[1:]
    parser = ArgumentParser(prog=prog, description='Mount Nintendo 3DS CTR Cart Image files.',
                            parents=(_c.default_argp, _c.ctrcrypto_argp, _c.main_args('cci', 'CCI file')))
    parser.add_argument('--dec', help='assume contents are decrypted', action='store_true')

    a = parser.parse_args(args)
    opts = dict(_c.parse_fuse_opts(a.o))

    if a.do:
        logging.basicConfig(level=logging.DEBUG, filename=a.do)

    cci_stat = get_time(a.cci)

    load_custom_boot9(a.boot9)

    with CCIReader(a.cci, dev=a.dev, assume_decrypted=a.dec) as r:
        mount = CTRCartImageMount(reader=r, g_stat=cci_stat)
        if _c.macos or _c.windows:
            opts['fstypename'] = 'CCI'
            if _c.macos:
                display = r.media_id.upper()
                try:
                    title = r.contents[CCISection.Application].exefs.icon.get_app_title()
                    display += f'; ' + r.contents[CCISection.Application].product_code
                    if title.short_desc != 'unknown':
                        display += '; ' + title.short_desc
                except:
                    pass
                opts['volname'] = f'CTR Cart Image ({display})'
            elif _c.windows:
                # volume label can only be up to 32 chars
                try:
                    title = r.contents[CCISection.Application].exefs.icon.get_app_title().short_desc
                    if len(title) > 26:
                        title = title[0:25] + '\u2026'  # ellipsis
                    display = title
                except:
                    display = r.tmd.title_id.upper()
                opts['volname'] = f'CCI ({display})'
        FUSE(mount, a.mount_point, foreground=a.fg or a.d, ro=True, nothreads=True, debug=a.d,
             fsname=realpath(a.cci).replace(',', '_'), **opts)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import tkinter as tk
import tkinter.ttk as ttk
from typing import TYPE_CHECKING

from .confighandler import get_bool, set_bool
from .optionsframes import CheckbuttonContainer

if TYPE_CHECKING:
    from . import NinfsGUI

# update options
CHECK_ONLINE = 'Check for updates on GitHub'


class NinfsSettings(tk.Toplevel):
    def __init__(self, parent: 'NinfsGUI'):
        super().__init__(parent)
        self.parent = parent

        self.wm_withdraw()
        self.parent.set_icon(self)
        self.wm_transient(self.parent)
        self.grab_set()
        self.wm_title('Settings')

        outer_container = ttk.Frame(self)
        outer_container.pack()

        update_check_frame = ttk.LabelFrame(outer_container, text='Updates')
        update_check_frame.pack(padx=10, pady=10)

        update_options = [CHECK_ONLINE]
        enabled = []
        if get_bool('update', 'onlinecheck'):
            enabled.append(CHECK_ONLINE)
        self.update_options_frame = CheckbuttonContainer(update_check_frame, options=update_options, enabled=enabled)
        self.update_options_frame.pack(padx=5, pady=5)

        footer_buttons = ttk.Frame(outer_container)
        footer_buttons.pack(padx=10, pady=(0, 10), side=tk.RIGHT)

        ok_button = ttk.Button(footer_buttons, text='OK', command=self.ok)
        ok_button.pack(side=tk.RIGHT)

        cancel_button = ttk.Button(footer_buttons, text='Cancel', command=self.cancel)
        cancel_button.pack(side=tk.RIGHT)

        self.wm_protocol('WM_DELETE_WINDOW', self.cancel)

        self.geometry("+%d+%d" % (parent.winfo_rootx() + 50, parent.winfo_rooty() + 50))

        self.wm_deiconify()

    def cancel(self):
        self.destroy()

    def ok(self):
        values = self.update_options_frame.get_values()
        set_bool('update', 'onlinecheck', values[CHECK_ONLINE])
        self.destroy()



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import tkinter as tk
from typing import TYPE_CHECKING

from .base import WizardBase
from .. import supportfiles

if TYPE_CHECKING:
    from .. import WizardContainer


class CTRNandImageSetup(WizardBase):
    def __init__(self, parent: 'tk.BaseWidget' = None, *, wizardcontainer: 'WizardContainer'):
        super().__init__(parent, wizardcontainer=wizardcontainer)

        def callback(*_):
            main_file = self.main_textbox_var.get().strip()
            b9_file = self.b9_textbox_var.get().strip()
            self.wizardcontainer.set_next_enabled(main_file and b9_file)

        main_container, main_textbox, main_textbox_var = self.make_file_picker('Select the NAND file:',
                                                                               'Select NAND file')
        main_container.pack(fill=tk.X, expand=True)

        b9_container, b9_textbox, b9_textbox_var = self.make_file_picker('Select the boot9 file:', 'Select boot9 file')
        b9_container.pack(fill=tk.X, expand=True)

        labeltext = 'Select the OTP file (not required if GodMode9 essentials.exefs is embedded):'
        otp_container, otp_textbox, otp_textbox_var = self.make_file_picker(labeltext, 'Select OTP file')
        otp_container.pack(fill=tk.X, expand=True)

        options_frame, cb_container = self.make_checkbox_options('Options:', ['Allow writing'])
        options_frame.pack(fill=tk.X, expand=True)
        self.cb_container = cb_container

        self.main_textbox_var = main_textbox_var
        self.b9_textbox_var = b9_textbox_var
        self.otp_textbox_var = otp_textbox_var

        main_textbox_var.trace_add('write', callback)
        b9_textbox_var.trace_add('write', callback)

        b9_textbox_var.set(supportfiles.last_b9_file)

        self.set_header_suffix('Nintendo 3DS NAND')

    def next_pressed(self):
        main_file = self.main_textbox_var.get().strip()
        b9_file = self.b9_textbox_var.get().strip()
        otp_file = self.otp_textbox_var.get().strip()

        if b9_file:
            supportfiles.last_b9_file = b9_file

        args = ['nandctr', main_file]
        opts = self.cb_container.get_values()
        if not opts['Allow writing']:
            args.append('-r')
        if b9_file:
            args += ['--boot9', b9_file]
        if otp_file:
            args += ['--otp', otp_file]

        self.wizardcontainer.show_mount_point_selector('3DS NAND', args)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import tkinter as tk
from typing import TYPE_CHECKING

from .base import WizardBase
from .. import supportfiles

if TYPE_CHECKING:
    from .. import WizardContainer


class SDFilesystemSetup(WizardBase):
    def __init__(self, parent: 'tk.BaseWidget' = None, *, wizardcontainer: 'WizardContainer'):
        super().__init__(parent, wizardcontainer=wizardcontainer)

        def callback(*_):
            main_file = self.main_textbox_var.get().strip()
            b9_file = self.b9_textbox_var.get().strip()
            movable_file = self.movable_textbox_var.get().strip()
            self.wizardcontainer.set_next_enabled(main_file and b9_file and movable_file)

        labeltext = 'Select the "Nintendo 3DS" directory:'
        main_container, main_textbox, main_textbox_var = self.make_directory_picker(labeltext,
                                                                                    'Select "Nintendo 3DS" directory')
        main_container.pack(fill=tk.X, expand=True)

        b9_container, b9_textbox, b9_textbox_var = self.make_file_picker('Select the boot9 file:', 'Select boot9 file')
        b9_container.pack(fill=tk.X, expand=True)

        movable_container, movable_textbox, movable_textbox_var = self.make_file_picker('Select the movable.sed file:',
                                                                                        'Select movable.sed file')
        movable_container.pack(fill=tk.X, expand=True)

        options_container, options_frame = self.make_checkbox_options('Options:', ['Allow writing'])
        options_container.pack(fill=tk.X, expand=True)

        self.main_textbox_var = main_textbox_var
        self.b9_textbox_var = b9_textbox_var
        self.movable_textbox_var = movable_textbox_var
        self.options_frame = options_frame

        main_textbox_var.trace_add('write', callback)
        b9_textbox_var.trace_add('write', callback)
        movable_textbox_var.trace_add('write', callback)

        b9_textbox_var.set(supportfiles.last_b9_file)

        self.set_header_suffix('Nintendo 3DS SD Card')

    def next_pressed(self):
        main_file = self.main_textbox_var.get().strip()
        b9_file = self.b9_textbox_var.get().strip()
        movable_file = self.movable_textbox_var.get().strip()

        if b9_file:
            supportfiles.last_b9_file = b9_file

        args = ['sd', main_file]
        opts = self.options_frame.get_values()
        if not opts['Allow writing']:
            args.append('-r')
        if b9_file:
            args += ['--boot9', b9_file]
        if movable_file:
            args += ['--movable', movable_file]

        self.wizardcontainer.show_mount_point_selector('3DS SD', args)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

"""
Mounts iQue Player NAND images, creating a virtual filesystem of the files contained within.
"""

import logging
from errno import ENOENT, EROFS
from stat import S_IFDIR, S_IFREG
from sys import exit, argv
from typing import BinaryIO

from pyctr.util import readbe

from . import _common as _c
# _common imports these from fusepy, and prints an error if it fails; this allows less duplicated code
from ._common import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context, get_time, realpath, basename

class BBNandImageMount(LoggingMixIn, Operations):
    fd = 0
    
    def __init__(self, nand_fp: BinaryIO, g_stat: dict):
        self.g_stat = g_stat
        
        self.files = {}
        
        self.f = nand_fp
        
    def __del__(self, *args):
        try:
            self.f.close()
        except AttributeError:
            pass
    
    destroy = __del__
    
    def init(self, path):
        nand_size = self.f.seek(0, 2)
        if nand_size != 0x4000000:
            exit(f'NAND size is incorrect (expected 0x4000000, got {nand_size:#X})')
        
        bbfs_blocks = []
        self.f.seek(0xFF0 * 0x4000)
        for i in range(0x10):
            bbfs_blocks.append(self.f.read(0x4000))
        
        self.f.seek(0)
        
        latest_seqno = -1
        latest_bbfs_block = None
        
        for i, j in enumerate(bbfs_blocks):
            header = j[0x3FF4:]
            
            magic = header[:4]
            if magic == b"\0x00\0x00\0x00\0x00":
                continue
            if magic not in [b"BBFS", b"BBFL"]:
                exit(f'Invalid BBFS magic: expected b"BBFS" or b"BBFL", got {magic.hex().upper()}')
            
            calculated_checksum = 0
            for k in range(0, 0x4000, 2):
                calculated_checksum += readbe(j[k:k + 2])
            
            if calculated_checksum & 0xFFFF != 0xCAD7:
                exit(f'BBFS block {i} has an invalid checksum')
            
            seqno = readbe(header[4:8])
            if seqno > latest_seqno:
                latest_seqno = seqno
                latest_bbfs_block = i
        
        if latest_bbfs_block == None or latest_seqno == -1:
            exit(f'Blank BBFS (all BBFS magics were 00000000)')
        
        self.used = 0
        
        for i in range(0x2000, 0x3FF4, 0x14):
            entry = bbfs_blocks[latest_bbfs_block][i:i + 0x14]
            valid = bool(entry[11])
            u = readbe(entry[12:14])
            start = u - (u & 0x8000) * 2
            if valid and start != -1:
                name = entry[:8].decode().rstrip("\x00")
                ext = entry[8:11].decode().rstrip("\x00")
                size = readbe(entry[16:20])
                self.files[f'/{name}.{ext}'] = {'start': start, 'size': size}
                self.used += size // 0x4000
        
        fat = bbfs_blocks[latest_bbfs_block][:0x2000]
        
        self.fat_entries = []
        for i in range(0, len(fat), 2):
            u = readbe(fat[i:i + 2])
            s = u - (u & 0x8000) * 2
            self.fat_entries.append(s)
    
    def flush(self, path, fh):
        return self.f.flush()
    
    @_c.ensure_lower_path
    def getattr(self, path, fh=None):
        uid, gid, pid = fuse_get_context()
        if path == '/':
            st = {'st_mode': (S_IFDIR | 0o777), 'st_nlink': 2}
        elif path in self.files:
            st = {'st_mode': (S_IFREG | 0o666),
                  'st_size': self.files[path]['size'], 'st_nlink': 1}
        else:
            raise FuseOSError(ENOENT)
        return {**st, **self.g_stat, 'st_uid': uid, 'st_gid': gid}
    
    def open(self, path, flags):
        self.fd += 1
        return self.fd
    
    @_c.ensure_lower_path
    def readdir(self, path, fh):
        yield from ('.', '..')
        yield from (x[1:] for x in self.files)
    
    @_c.ensure_lower_path
    def read(self, path, size, offset, fh):
        fi = self.files[path]
        
        if offset > fi['size']:
            return b''
        
        data = bytearray()
        
        block = fi['start']
        while True:
            self.f.seek(block * 0x4000)
            data.extend(self.f.read(0x4000))
            block = self.fat_entries[block]
            if block == -1:
                break
            if block in [0, -2, -3]:
                return b''
        
        if len(data) != fi['size']:
            return b''
        
        if offset + size > fi['size']:
            size = fi['size'] - offset
        
        return bytes(data[offset:offset + size])
    
    @_c.ensure_lower_path
    def statfs(self, path):
        return {'f_bsize': 0x4000, 'f_frsize': 0x4000, 'f_blocks': 0xFF0 - 0x40, 'f_bavail': 0xFF0 - 0x40 - self.used,
                'f_bfree': 0xFF0 - 0x40 - self.used, 'f_files': len(self.files)}
    
    @_c.ensure_lower_path
    def write(self, path, data, offset, fh):
        raise FuseOSError(EROFS)


def main(prog: str = None, args: list = None):
    from argparse import ArgumentParser
    if args is None:
        args = argv[1:]
    parser = ArgumentParser(prog=prog, description='Mount iQue Player NAND images.',
                            parents=(_c.default_argp, _c.main_args('nand', 'iQue Player NAND image')))
    
    a = parser.parse_args(args)
    opts = dict(_c.parse_fuse_opts(a.o))
    
    if a.do:
        logging.basicConfig(level=logging.DEBUG, filename=a.do)
    
    nand_stat = get_time(a.nand)
    
    with open(a.nand, 'rb') as f:
        mount = BBNandImageMount(nand_fp=f, g_stat=nand_stat)
        if _c.macos or _c.windows:
            opts['fstypename'] = 'BBFS'
            if _c.macos:
                opts['volname'] = f'iQue Player NAND ({basename(a.nand)})'
            elif _c.windows:
                # volume label can only be up to 32 chars
                opts['volname'] = 'iQue Player NAND'
        FUSE(mount, a.mount_point, foreground=a.fg or a.d, ro=True, nothreads=True, debug=a.d,
             fsname=realpath(a.nand).replace(',', '_'), **opts)



#!/usr/bin/env python3

import sys

from setuptools import setup, find_packages

from ninfs import mountinfo

if sys.hexversion < 0x030800f0:
    sys.exit('Python 3.8.0+ is required.')

with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

# based on https://github.com/Legrandin/pycryptodome/blob/b3a394d0837ff92919d35d01de9952b8809e802d/setup.py
with open('ninfs/__init__.py', 'r', encoding='utf-8') as f:
    for line in f:
        if line.startswith('__version__'):
            version = eval(line.split('=')[1])

setup(
    name='ninfs',
    version=version,
    packages=find_packages(),
    url='https://github.com/ihaveamac/ninfs',
    license='MIT',
    author='Ian Burgwin',
    author_email='ian@ianburgwin.net',
    description='FUSE filesystem Python scripts for Nintendo console files',
    long_description=readme,
    long_description_content_type='text/markdown',
    package_data={'ninfs.gui': ['data/*.png', 'data/*.ico', 'data/licenses/*']},
    classifiers=[
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    install_requires=['pycryptodomex>=3.9,<4', 'pyctr>=0.7,<0.8', 'haccrypto>=0.1', 'pypng>=0.0.21'],
    python_requires='>=3.8.0',
    # fusepy should be added here once the main repo has a new release with Windows support.
    entry_points={'gui_scripts': ['ninfsw = ninfs.main:gui'],
                  'console_scripts': ['ninfs = ninfs.main:gui'] +
                                     [f'mount_{x} = ninfs.main:main' for x in mountinfo.types.keys()] +
                                     [f'mount_{x} = ninfs.main:main' for x in mountinfo.aliases.keys()]}
)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import tkinter as tk
from typing import TYPE_CHECKING

from .base import WizardBase

if TYPE_CHECKING:
    from .. import WizardContainer


class HACNandImageSetup(WizardBase):
    def __init__(self, parent: 'tk.BaseWidget' = None, *, wizardcontainer: 'WizardContainer'):
        super().__init__(parent, wizardcontainer=wizardcontainer)

        def callback(*_):
            main_file = self.main_textbox_var.get().strip()
            keys_file = self.keys_textbox_var.get().strip()
            self.wizardcontainer.set_next_enabled(main_file and keys_file)

        filetypes = [
            'Full NAND image',
            'PRODINFO',
            'PRODINFOF',
            'SAFE',
            'SYSTEM',
            'USER',
        ]
        filetype_container, filetype_menu, filetype_menu_var = self.make_option_menu('Select the file type:',
                                                                                     *filetypes)
        filetype_container.pack(fill=tk.X, expand=True)

        labeltext = 'Select the NAND file (full or split):'
        main_container, main_textbox, main_textbox_var = self.make_file_picker(labeltext, 'Select NAND file')
        main_container.pack(fill=tk.X, expand=True)

        keys_container, keys_textbox, keys_textbox_var = self.make_file_picker('Select the BIS keys file:',
                                                                               'Select BIS keys file')
        keys_container.pack(fill=tk.X, expand=True)

        options_frame, cb_container = self.make_checkbox_options('Options:', ['Allow writing'])
        options_frame.pack(fill=tk.X, expand=True)
        self.cb_container = cb_container

        self.main_textbox_var = main_textbox_var
        self.keys_textbox_var = keys_textbox_var
        self.filetype_var = filetype_menu_var

        main_textbox_var.trace_add('write', callback)
        keys_textbox_var.trace_add('write', callback)

        self.set_header_suffix('Nintendo Switch NAND')

    def next_pressed(self):
        main_file = self.main_textbox_var.get().strip()
        keys_file = self.keys_textbox_var.get().strip()
        filetype = self.filetype_var.get()

        args = ['nandhac', main_file]
        if main_file[-3] == '.':
            try:
                int(main_file[-2:])
            except ValueError:
                # not a split file
                pass
            else:
                # is a split file
                args.append('--split-files')
        opts = self.cb_container.get_values()
        if not opts['Allow writing']:
            args.append('-r')
        if keys_file:
            args += ['--keys', keys_file]
        if filetype != 'Full NAND image':
            args += ['--partition', filetype]

        self.wizardcontainer.show_mount_point_selector('Switch NAND', args)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.
import os
import sys
import webbrowser
from inspect import cleandoc
from os.path import dirname, join
from pprint import pformat
from subprocess import Popen, PIPE, STDOUT, TimeoutExpired, check_call
from threading import Thread
from traceback import format_exc
from typing import TYPE_CHECKING
from uuid import uuid4

is_windows = sys.platform == 'win32'
is_mac = sys.platform == 'darwin'

try:
    import tkinter as tk
    import tkinter.ttk as ttk
    import tkinter.messagebox as mb
except ModuleNotFoundError:
    print('Could not find tkinter.', file=sys.stderr)
    if is_mac:
        print('If using a package manager like Homebrew or MacPorts, make sure to install Tkinter with it.',
              file=sys.stderr)
    elif not is_mac or is_windows:
        print('Make sure to install Tkinter with your package manager.', file=sys.stderr)
    sys.exit(1)

from .about import NinfsAbout
from .confighandler import get_bool, set_bool
from .settings import NinfsSettings
from .updatecheck import thread_update_check
from .wizardcontainer import WizardContainer, WizardTypeSelector, WizardFailedMount

if TYPE_CHECKING:
    from typing import Callable, Dict, List, Tuple

tutorial_url = 'https://gbatemp.net/threads/499994/'

# cx_Freeze, PyInstaller, etc.
frozen = getattr(sys, 'frozen', None)
# PyInstaller sets this
meipass = getattr(sys, '_MEIPASS', None)

executable = sys.executable
if is_windows:
    from os.path import isdir as check_mountpoint
    from signal import CTRL_BREAK_EVENT
    from subprocess import CREATE_NEW_PROCESS_GROUP

    if frozen:
        print('Using console exe')
        executable = join(dirname(sys.executable), 'ninfs.exe')
        print(executable)
else:
    from os.path import ismount as check_mountpoint


def thread_output_reader(gui: 'NinfsGUI', proc: 'Popen', uuid: 'str', output_list: 'List[str]'):
    while proc.poll() is None:
        for line in proc.stdout:
            if line != '':
                line = line.rstrip('\r\n')
                output_list.append(line)

    # if the uuid is not in the mounts dict, then it was killed by this script
    if proc.returncode and uuid in gui.mounts:
        gui.remove_mount_info(uuid)
        wizard_window = WizardContainer(gui)
        wizard_window.change_frame(WizardFailedMount, returncode=proc.returncode, output=output_list, kind='crash')
        wizard_window.focus()


class NinfsGUI(tk.Tk):
    def __init__(self):
        super().__init__()

        self.mounts: Dict[str, Tuple[Popen, Thread, List[str], str]] = {}

        container = ttk.Frame(self)
        container.pack(fill=tk.BOTH, expand=True)

        self.wm_withdraw()

        self.wm_title('ninfs')

        self.ico_path = self.get_data_file('windows.ico')

        self.wm_minsize(500, 300)
        self.create_menu_bar()

        style = ttk.Style(container)
        style.configure('TMenubutton', background='gainsboro')

        container.rowconfigure(0, weight=0)
        container.rowconfigure(1, weight=1)
        container.rowconfigure(2, weight=0)
        container.columnconfigure(0, weight=1)

        header = ttk.Label(container, text='Mounted contents', font=(None, 15, 'bold'), justify=tk.LEFT)
        header.grid(row=0, column=0, padx=10, pady=8, sticky=tk.W)

        mount_treeview_frame = ttk.Frame(container)
        mount_treeview_frame.grid(row=1, column=0, sticky=tk.NSEW, padx=10)
        mount_treeview_frame.rowconfigure(0, weight=1)
        mount_treeview_frame.columnconfigure(0, weight=1)
        mount_treeview_frame.columnconfigure(1, weight=0)

        self.mount_treeview = ttk.Treeview(mount_treeview_frame)
        self.mount_treeview.grid(row=0, column=0, sticky=tk.NSEW)
        self.mount_treeview.configure(columns=('mount_path', 'mount_type', 'mounted_item'), show='headings')

        self.mount_treeview.column('mount_path', width=100, anchor=tk.W)
        self.mount_treeview.heading('mount_path', text='Mount Path')
        self.mount_treeview.column('mount_type', width=50, anchor=tk.W)
        self.mount_treeview.heading('mount_type', text='Type')
        self.mount_treeview.column('mounted_item', width=200, anchor=tk.W)
        self.mount_treeview.heading('mounted_item', text='Mounted Content')

        mount_treeview_scrollbar = ttk.Scrollbar(mount_treeview_frame, orient=tk.VERTICAL,
                                                 command=self.mount_treeview.yview)
        self.mount_treeview.configure(yscrollcommand=mount_treeview_scrollbar.set)
        mount_treeview_scrollbar.grid(row=0, column=1, sticky=tk.NS)

        actions_frame = ttk.Frame(container)
        actions_frame.grid(row=2, column=0, padx=10, pady=10, sticky=tk.W)

        new_mount_button = ttk.Button(actions_frame, text='New mount', command=self.show_wizard)
        new_mount_button.pack(side=tk.LEFT)

        unmount = ttk.Button(actions_frame, text='Unmount selected', command=self.unmount_selected)
        unmount.pack(side=tk.LEFT)

        self.wm_protocol('WM_DELETE_WINDOW', self.on_close)

    def get_data_file(self, path):
        if frozen and meipass:
            return join(meipass, 'guidata', path)
        else:
            # this kinda sucks
            return join(dirname(__file__), 'data', path)

    def set_icon(self, wm: 'tk.Wm'):
        # just in case this also works on other platforms, but this assumption is likely wrong lol
        # TODO: fix this for Linux
        if is_windows:
            wm.wm_iconbitmap(self.ico_path)

    def check_fuse(self):
        try:
            import fuse
        except EnvironmentError as e:
            if e.args[0] == 'Unable to find libfuse':
                if is_windows:
                    res = mb.askyesno('Failed to load libfuse',
                                      'Failed to load libfuse. WinFsp needs to be installed.\n\n'
                                      'Would you like to open the WinFsp download page?\n'
                                      'https://winfsp.dev/rel/')
                    if res:
                        webbrowser.open('https://winfsp.dev/rel/')
                elif is_mac:
                    res = mb.askyesno('Failed to load libfuse',
                                      'Failed to load libfuse. macFUSE needs to be installed.\n\n'
                                      'Would you like to open the macFUSE download page?\n'
                                      'https://osxfuse.github.io')
                    if res:
                        webbrowser.open('https://osxfuse.github.io')
                else:
                    mb.showerror('Failed to load libfuse.')
                return False
        except Exception:
            mb.showerror('Failed to import fusepy', 'Failed to import fusepy.\n\n' + format_exc())
            return False
        return True

    def mainloop(self, n=0):
        self.wm_deiconify()

        if not get_bool('internal', 'askedonlinecheck'):
            message = '''
            Check for updates online?
            This will make a request to GitHub every time the ninfs gui is opened.
            
            This can be changed any time in Settings.
            '''
            if mb.askyesno('Check for updates', cleandoc(message)):
                set_bool('update', 'onlinecheck', True)
            set_bool('internal', 'askedonlinecheck', True)

        if get_bool('update', 'onlinecheck'):
            update_thread = Thread(target=thread_update_check, args=(self,))
            update_thread.start()
        super().mainloop(n)

    def on_close(self):
        if self.mounts:
            if mb.askokcancel('Exiting', 'Closing ninfs will unmount all contents.'):
                self.unmount_all(force=True)
                self.destroy()
        else:
            self.destroy()

    def show_settings(self):
        settings_window = NinfsSettings(self)
        settings_window.focus()

    def show_wizard(self):
        wizard_window = WizardContainer(self)
        wizard_window.change_frame(WizardTypeSelector)
        wizard_window.focus()

    def mount(self, mounttype: str, cmdargs: 'List[str]', mountpoint: str, callback_success: 'Callable',
              callback_failed: 'Callable', is_drive_letter: bool):
        if is_windows:
            if not is_drive_letter:
                # the directory must first be deleted before winfsp can mount to it
                try:
                    os.rmdir(mountpoint)
                except FileNotFoundError:
                    pass
                except Exception:
                    exc_list = ['Failed to delete directory:', mountpoint, '']
                    exc_list.extend(format_exc().splitlines())
                    callback_failed(None, exc_list)
                    return

        args = [executable]
        if not frozen:
            args.append(dirname(dirname(__file__)))
        args.extend(cmdargs)
        args.append('-f')
        args.append(mountpoint)

        popen_opts = {}
        if is_windows:
            popen_opts['creationflags'] = CREATE_NEW_PROCESS_GROUP

        uuid = str(uuid4())

        output_list = ['Command: ' + pformat(args), '']
        proc = Popen(args, stdin=PIPE, stdout=PIPE, stderr=STDOUT, encoding='utf-8', **popen_opts)
        thread = Thread(target=thread_output_reader, args=(self, proc, uuid, output_list))

        mount_info = (proc, thread, output_list, mountpoint)

        def check_loop():
            if check_mountpoint(mountpoint):
                self.mount_treeview.insert('', tk.END, text=uuid, iid=uuid, values=(mountpoint, mounttype, cmdargs[1]))
                self.mounts[uuid] = mount_info
                callback_success()
                return
            if proc.poll() is not None:
                thread.join()
                callback_failed(proc.returncode, output_list)
                return

            self.after(500, check_loop)

        thread.start()

        self.after(500, check_loop)

    def unmount_selected(self):
        # the mounts dict gets modified during iteration, so the list of keys is cloned to prevent issues
        # it's also done in reverse, since later mounts might be based on earlier ones
        selection = self.mount_treeview.selection()
        for s in reversed(selection):
            self.unmount(s)

    def unmount_all(self, *, force: bool = False):
        # the mounts dict gets modified during iteration, so the list of keys is cloned to prevent issues
        # it's also done in reverse, since later mounts might be based on earlier ones
        for uuid in reversed(self.mount_treeview.get_children()):
            self.unmount(uuid, force=force)

    def remove_mount_info(self, uuid: str):
        self.mount_treeview.delete(uuid)
        del self.mounts[uuid]

    def unmount(self, uuid: 'str', *, force: bool = False):
        mount_info = self.mounts[uuid]
        if is_windows:
            mount_info[0].send_signal(CTRL_BREAK_EVENT)
            try:
                mount_info[0].wait(3)
            except TimeoutExpired:
                if force:
                    self.remove_mount_info(uuid)
                    mount_info[0].kill()
                else:
                    res = mb.askyesno('Mount not responding', 'The mount subprocess is not responding.\nTerminate it?')
                    if res:
                        self.remove_mount_info(uuid)
                        mount_info[0].kill()
            else:
                self.remove_mount_info(uuid)
        else:
            # I think this is cheating
            if is_mac:
                check_call(['diskutil', 'unmount', mount_info[3]])
                self.remove_mount_info(uuid)
            else:
                # assuming linux or bsd, which have fusermount
                check_call(['fusermount', '-u', mount_info[3]])
                self.remove_mount_info(uuid)

    @staticmethod
    def show_tutorial():
        webbrowser.open(tutorial_url)

    def show_about(self):
        about_window = NinfsAbout(self)
        about_window.focus()

    def create_menu_bar(self):
        self.option_add('*tearOff', tk.FALSE)
        menubar = tk.Menu(self)

        if is_mac:
            apple_menu = tk.Menu(menubar, name='apple')
            apple_menu.add_command(label='About ninfs', command=self.show_about)
            apple_menu.add_separator()
            menubar.add_cascade(menu=apple_menu)

            self.createcommand('tk::mac::ShowPreferences', self.show_settings)
            self.createcommand('tk::mac::Quit', self.on_close)

        file_menu = tk.Menu(menubar)
        if not is_mac:
            file_menu.add_command(label='Settings', command=self.show_settings)

        help_menu = tk.Menu(menubar)
        help_menu.add_command(label='Open tutorial on GBAtemp', command=self.show_tutorial)
        if not is_mac:
            help_menu.add_command(label='About ninfs', command=self.show_about)

        menubar.add_cascade(label='File', menu=file_menu)
        menubar.add_cascade(label='Help', menu=help_menu)

        self.configure(menu=menubar)


def start_gui():
    window = NinfsGUI()

    if is_windows:
        from ctypes import windll, get_last_error
        from os import environ, getpid

        environ['NINFS_GUI_PARENT_PID'] = str(getpid())

        if not windll.kernel32.GetConsoleWindow():
            # if there is no console, make one and hide it
            # this is not an elegant solution but it lets us use send_signal on subprocesses
            if not windll.kernel32.AllocConsole():
                # AllocConsole fails when I'm testing in PyCharm but get_last_error returns 0, meaning it succeeded.
                # I don't know why this happens.
                err = get_last_error()
                if err:
                    print('Failed to use AllocConsole:', err)
            else:
                windll.user32.ShowWindow(windll.kernel32.GetConsoleWindow(), 0)  # SW_HIDE

    if window.check_fuse():
        window.mainloop()
        return 0
    else:
        return 70
